import copy
from abc import ABC, abstractmethod
from typing import List

import numpy as np
from ase import Atoms
from ase.io import write as ase_write
from scipy.spatial import ConvexHull, Delaunay

from ..generator import AmorphousGenerator, BaseGenerator
from ..transform import BaseTransform
from .algebraic import BaseAlgebraic, Plane, Sphere
from .algebraic import get_bounding_box as get_bounding_box_planes

############################
###### Volume Classes ######
############################


class BaseVolume(ABC):
    """Base abstract class for Volume objects.
    
    Volume objects are subtractive components in Construction Zone. When designing
    nanostructures, Volumes contain information about where atoms should and
    should not be placed. Semantically, volumes can be thought of as singular
    objects in space.

    BaseVolumes are typically not created directly. Use the Volume class for
    generalized convex objects, and the MultiVolume class for unions of convex
    objects.

    Attributes:
        atoms (np.ndarray): Nx3 array of atom positions of atoms lying within volume.
        species (np.ndarray): Nx1 array of atomic numbers of atoms lying within volume.
        ase_atoms (Atoms): Collection of atoms in volume as ASE Atoms object.
        priority (int): Relative generation precedence of volume.
    """

    @abstractmethod
    def __init__(self, **kwargs):
        pass

    @property
    def atoms(self):
        """Array of atomic positions of atoms lying within volume."""
        return self._atoms

    @property
    def species(self):
        """Array of atomic numbers of atoms lying within volume."""
        return self._species

    @property
    def ase_atoms(self):
        """Collection of atoms in volume as ASE Atoms object."""
        return Atoms(symbols=self.species, positions=self.atoms)

    @property
    def priority(self):
        """Relative generation precedence of volume."""
        return self._priority

    @priority.setter
    def priority(self, priority):
        if not isinstance(priority, int):
            raise TypeError("Priority needs to be integer valued")

        self._priority = priority

    @abstractmethod
    def transform(self, transformation):
        """Transform volume with given transformation.

        Args:
            transformation (BaseTransform): transformation to apply to volume.
        """
        pass

    @abstractmethod
    def populate_atoms(self):
        """Fill volume with atoms. """
        pass

    @abstractmethod
    def checkIfInterior(self, testPoints: np.ndarray):
        """Check points to see if they lie in interior of volume.
        
        Returns:
            Logical array indicating which points lie inside the volume.
        """
        pass

    def to_file(self, fname, **kwargs):
        """Write object to an output file, using ASE write utilities.

        Args:
            fname (str): output file name.
            **kwargs: any key word arguments otherwise accepted by ASE write.
        """
        ase_write(filename=fname, images=self.ase_atoms, **kwargs)


class Volume(BaseVolume):
    """Volume object for representing convex spaces.
    
    Volume objects are subtractive components in Construction Zone. When designing
    nanostructures, Volumes contain information about where atoms should and
    should not be placed. Semantically, volumes can be thought of as singular
    objects in space. In order to supply atoms, Volumes must be given a Generator.

    Volumes can be created with a series of points in space, in which the 
    interior of the volume is taken as the convex hull of the points in space.
    They can also be created with a series of algebraic surfaces, such as planes
    and spheres. Both points and algebraic objects can be used to define a Volume,
    in which the interior of the Volume is taken as the intersection of the 
    interior region defined by the convex hull of the points and the interior
    regions of the algebraic objects.

    Attributes:
        points (np.ndarray): Nx3 array of points used to defined convex hull.
        alg_objects (List[BaseAlgebraic]): Algebraic objects used to define convex region.
        hull (ConvexHull): Convex hull of points defining volume.
        tri (Delaunay): Delaunay triangulation of facets of convex hull.
        generator (Generator): Generator object associated with volume that supplies atoms.
        atoms (np.ndarray): Nx3 array of atom positions of atoms lying within volume.
        species (np.ndarray): Nx1 array of atomic numbers of atoms lying within volume.
        ase_atoms (Atoms): Collection of atoms in volume as ASE Atoms object.
        priority (int): Relative generation precedence of volume.
    """

    def __init__(self,
                 points: np.ndarray = None,
                 alg_objects: np.ndarray = None,
                 generator: BaseGenerator = None,
                 priority: int = 0,
                 **kwargs):
        self._points = None
        self._hull = None
        self._generator = None
        self._atoms = None
        self._tri = None
        self._alg_objects = []
        self._priority = 0

        if not (points is None):
            #expect 2D array with Nx3 points
            assert (len(
                points.shape) == 2), "points must be N x 3 numpy array (x,y,z)"
            assert (points.shape[1] == 3
                   ), "points must be N x 3 numpy array (x,y,z)"
            self.addPoints(points)

        if not (generator is None):
            if 'gen_origin' in kwargs:
                self.add_generator(generator, kwargs["gen_origin"])
            else:
                self.add_generator(generator)

        self.priority = priority

        if not (alg_objects is None):
            for obj in alg_objects:
                self.add_alg_object(obj)

    """
    Properties
    """

    @property
    def points(self):
        """Nx3 array of points used to defined convex hull."""
        return self._points

    @points.setter
    def points(self, points):
        try:
            self._points = np.array([])  #clear points
            self.addPoints(points)
        except AssertionError:
            raise ValueError("Check shape of input array.")

    @property
    def alg_objects(self):
        """Algebraic objects used to define convex region."""
        return self._alg_objects

    def add_alg_object(self, obj: BaseAlgebraic):
        """Add an algebraic surface to the volume.
        
        Args:
            obj (BaseAlgebraic): Algebraic surface to add to volume.
        """
        assert (
            isinstance(obj, (BaseAlgebraic))
        ), "Must be adding algebraic objects from derived BaseAlgebraic class"
        self._alg_objects.append(copy.deepcopy(obj))

    @property
    def hull(self):
        """Convex hull of points defining volume."""
        return self._hull

    @property
    def tri(self):
        """Delaunay triangulation of facets of convex hull."""
        return self._tri

    @property
    def generator(self):
        """Generator object associated with volume that supplies atoms."""
        return self._generator

    def add_generator(self, generator, origin=None):
        if not isinstance(generator, BaseGenerator):
            raise TypeError("Supplied generator is not of Generator() class")

        new_generator = copy.deepcopy(generator)

        if not isinstance(generator, AmorphousGenerator):
            if not origin is None:
                new_generator.voxel.origin = origin

        self._generator = new_generator

    """
    Methods
    """

    def createHull(self):
        """Create convex hull from points defining volume boundaries."""
        #check to make sure there are N>3 points in point list
        assert (self.points.shape[0] >
                3), "must have more than 3 points to create hull"
        self._hull = ConvexHull(self.points, incremental=True)
        self._tri = Delaunay(self.hull.points[self.hull.vertices])

    def addPoints(self, points: np.ndarray):
        """Add points to list of points defining convex hull and update hull.

        Args:
            points (np.ndarray): Nx3 array of points to add to hull.
        """
        assert (
            points.shape[-1] == 3), "points must be N x 3 numpy array (x,y,z)"
        assert (len(points.shape) <
                3), "points must be N x 3 numpy array (x,y,z)"

        if (self._points is None):
            self._points = np.copy(points)
            if len(points.shape) == 1:  #add dim if only single new point
                self._points = np.expand_dims(points)
        else:
            if len(points.shape) == 1:  #check for single point
                points = np.expand_dims(points, axis=0)

            self._points = np.append(self._points, points, axis=0)

        #if hull created, update points; else, create hull
        try:
            self._hull.add_points(points)
        except AttributeError:
            self.createHull()

    def transform(self, transformation: BaseTransform):

        assert (isinstance(transformation, BaseTransform)
               ), "Supplied transformation not transformation object."

        if not (self.points is None):
            self.points = transformation.applyTransformation(self.points)
            self.createHull()

        if len(self.alg_objects) > 0:
            for i, obj in enumerate(self.alg_objects):
                self.alg_objects[i] = transformation.applyTransformation_alg(
                    obj)

        if transformation.locked and (not (self.generator is None)):
            self.generator.transform(transformation)

    def checkIfInterior(self, testPoints: np.ndarray):
        assert (testPoints.shape[-1] == 3
               ), "testPoints must be N x 3 numpy array (x,y,z)"
        assert (len(testPoints.shape) <
                3), "testPoints must be N x 3 numpy array (x,y,z)"
        if (len(testPoints.shape) == 1):
            testPoints = np.expand_dims(testPoints, axis=0)

        check = np.ones(testPoints.shape[0]).astype(bool)

        if not self.tri is None:
            check = np.logical_and(
                check,
                self.tri.find_simplex(testPoints, tol=2.5e-1) >= 0)

        if len(self.alg_objects) > 0:
            for obj in self.alg_objects:
                check = np.logical_and(check, obj.checkIfInterior(testPoints))

        return check

    def get_bounding_box(self):
        """Get some minimal bounding box defining extremities of regions.
        
        Returns:
            Nx3 array of points defining extremities of region enclosed by volume.
        """
        if not (self.points is None):
            return self.points
        else:
            # As heuristic, look for any sphere first
            # Then, gather planes and check if valid intersection exists
            spheres = [
                obj for obj in self.alg_objects if isinstance(obj, Sphere)
            ]
            if len(spheres) > 0:
                d = 2 * spheres[0].radius
                bbox = makeRectPrism(d, d, d)
                shift = spheres[0].center - (d / 2) * np.ones(3)
                return bbox + shift

            planes = [obj for obj in self.alg_objects if isinstance(obj, Plane)]
            if len(planes) > 3:
                return get_bounding_box_planes(planes)

    def populate_atoms(self):
        bbox = self.get_bounding_box()
        coords, species = self.generator.supply_atoms(bbox)
        check = self.checkIfInterior(coords)

        self._atoms = coords[check, :]
        self._species = species[check]

    def from_volume(self, **kwargs):
        """Constructor for new Volumes based on existing Volume object.

        Args:
            **kwargs: 
                    - transformation=List[BaseTransformation] to apply a series
                     of transfomrations to copied Volume.
                    - generator=BaseGenerator to replace generator associated with volume.
                    - Any kwargs accepted in creation of Volume object.
        """
        new_volume = Volume(points=self.points,
                            alg_objects=self.alg_objects,
                            priority=self.priority)
        if "generator" in kwargs.keys():
            new_volume.add_generator(kwargs["generator"])
        else:
            new_volume.add_generator(self.generator)

        if "transformation" in kwargs.keys():
            for t in kwargs["transformation"]:
                new_volume.transform(t)

        return new_volume


class MultiVolume(BaseVolume):
    """Volume object for representing arbitrary union of convex spaces.

    Volume objects are subtractive components in Construction Zone. When designing
    nanostructures, Volumes contain information about where atoms should and
    should not be placed. Semantically, volumes can be thought of as singular
    objects in space. In order to supply atoms, Volumes must be given a Generator.

    MultiVolumes group multiple Volume objects together into a single semantic object.
    Within the MultiVolume, Volume intersection is handled with relative precedence levels, 
    analagous to the precedence relationships that are used to handle conflict
    resolution between Volumes in scenes. Transformations applied to a MultiVolume
    are applied to every owned volume. MultiVolumes can be nested.

    Attributes:
        volumes (np.ndarray): Nx3 array of points used to defined convex hull.
        atoms (np.ndarray): Nx3 array of atom positions of atoms lying within volume.
        species (np.ndarray): Nx1 array of atomic numbers of atoms lying within volume.
        ase_atoms (Atoms): Collection of atoms in volume as ASE Atoms object.
        priority (int): Relative generation precedence of volume.
    """

    def __init__(self, volumes: List[BaseVolume] = None, priority: int = None):
        self._priority = 0
        self._volumes = []
        if not (volumes is None):
            self.add_volume(volumes)

        if not (priority is None):
            self.priority = priority

    @property
    def volumes(self):
        """Collection of volumes grouped in MultiVolume."""
        return self._volumes

    def add_volume(self, volume: BaseVolume):
        """Add volume to MultiVolume.
        
        Args:
            volume (BaseVolume): Volume object to add to MultiVolume.
        """
        if hasattr(volume, '__iter__'):
            for v in volume:
                assert (isinstance(
                    v, BaseVolume)), "volumes must be volume objects"
            self._volumes.extend(volume)
        else:
            assert (isinstance(volume,
                               BaseVolume)), "volumes must be volume objects"
            self._volumes.append(volume)

    def _get_priorities(self):
        """Grab priority levels of all volumes in MultiVolume to determine precedence relationship.

        Returns:
            List of relative priority levels and offsets. Relative priority levels
            and offsets are used to determine which objects whill be checked
            for the inclusion of atoms in the scene of the atoms contributed by
            another object.
        """

        # get all priority levels active first
        self.volumes.sort(key=lambda ob: ob.priority)
        plevels = np.array([x.priority for x in self.volumes])

        # get unique levels and create relative priority array
        __, idx = np.unique(plevels, return_index=True)
        rel_plevels = np.zeros(len(self.volumes)).astype(int)
        for i in idx[1:]:
            rel_plevels[i:] += 1

        offsets = np.append(idx, len(self.volumes))

        return rel_plevels, offsets

    def transform(self, transformation: BaseTransform):
        assert (isinstance(transformation, BaseTransform)
               ), "Supplied transformation not transformation object."

        for vol in self.volumes:
            vol.transform(transformation)

    def checkIfInterior(self, testPoints: np.ndarray):
        assert (testPoints.shape[-1] == 3
               ), "testPoints must be N x 3 numpy array (x,y,z)"
        assert (len(testPoints.shape) <
                3), "testPoints must be N x 3 numpy array (x,y,z)"
        if (len(testPoints.shape) == 1):
            testPoints = np.expand_dims(testPoints, axis=0)

        check = np.zeros(testPoints.shape[0]).astype(bool)

        for vol in self.volumes:
            check = np.logical_or(check, vol.checkIfInterior(testPoints))

        return check

    def populate_atoms(self):
        #routine is modified form of scene atom population
        for vol in self.volumes:
            vol.populate_atoms()

        rel_plevels, offsets = self._get_priorities()

        checks = []

        for i, vol in enumerate(self.volumes):
            check = np.ones(vol.atoms.shape[0]).astype(bool)
            eidx = offsets[rel_plevels[i] + 1]

            for j in range(eidx):
                if (i != j):
                    check_against = np.logical_not(
                        self.volumes[j].checkIfInterior(vol.atoms))
                    check = np.logical_and(check, check_against)

            checks.append(check)

        self._atoms = np.vstack(
            [vol.atoms[checks[i], :] for i, vol in enumerate(self.volumes)])
        self._species = np.hstack(
            [vol.species[checks[i]] for i, vol in enumerate(self.volumes)])

    def from_volume(self, **kwargs):
        """Constructor for new MultiVolume based on existing MultiVolume object.

        **kwargs passed to volume are applied to every owned Volume individually.

        Args:
            **kwargs: 
                    - transformation=List[BaseTransformation] to apply a series
                     of transfomrations to copied Volume.
                    - generator=BaseGenerator to replace generator associated with volume.
                    - Any kwargs accepted in creation of Volume object.
        """
        new_vols = []
        for vol in self.volumes:
            new_vols.append(vol.from_volume(**kwargs))

        return MultiVolume(volumes=new_vols, priority=self.priority)


############################
#### Utility functions #####
############################


def makeRectPrism(a, b, c, center=None):
    """Create rectangular prism.

    Args:
        a (float): dimension of prism along x
        b (float): dimension of prism along y
        c (float): dimension of prism along z
        center (np.ndarray): center of prism, default None. If None, corner of
                            prism is at origin. Else, prism is translated to 
                            have midpoint at center.

    Returns:
        8x3 numpy array of 8 points defining a rectangular prism in space.
    """
    points = np.array([[0, 0, 0], [1, 0, 0], [0, 1, 0], [0, 0, 1], [1, 1, 0],
                       [1, 0, 1], [0, 1, 1], [1, 1, 1]],
                      dtype=np.float64)
    #stretch unit cube
    points *= np.array([a, b, c])

    if center is None:
        return points
    else:
        #translate prism to desired center if specified
        cur_center = np.mean(points, axis=0)
        return points + (center - cur_center)



from typing import List, Union

import numpy as np
from numpy.typing import ArrayLike


def round_away(x: float) -> float:
    """Round to float integer away from zero--opposite of np.fix.
    
    Args:
        x (float, ArrayLike[float]): number(s) to round

    Returns:
        float, ArrayLike[float]: rounded number(s)
    """
    return np.sign(x) * np.ceil(np.abs(x))


def get_N_splits(n: int, m: int, l: int, seed: int = None) -> List[int]:
    """Get N uniform random integers in interval [M,L-M) with separation M.

    Args:
        n (int): number of indices
        m (int): minimum distance between indices and ends of list
        l (int): length of initial list
        seed (int): seed for random number generator, default None

    Returns:
        List[int]: sorted list of random indices
    """
    if (l - 2 * m < (n - 1) * m):
        raise ValueError("m is too large for number of splits requested and l")
    rng = np.random.default_rng(seed=seed)

    #seed an initial choice and create array to calculate distances in
    splits = [rng.integers(m, l - m)]
    data = np.array([x for x in range(m, l - m)])
    idx = np.ma.array(data=data, mask=np.abs(data - splits[-1]) < m)

    while (len(splits) < n):
        while (np.all(idx.mask)):
            # no options left, reseed
            splits = [rng.integers(m, l - m)]
            idx.mask = np.abs(idx.data - splits[-1]) < m

        # add new choice to list and check distance against other indices
        splits.append(rng.choice(idx.compressed()))
        idx.mask = np.logical_or(idx.mask, np.abs(idx.data - splits[-1]) < m)

    splits.sort()
    return splits


def vangle(v1: np.ndarray, v2: np.ndarray) -> float:
    """Calculate angle between two vectors of same dimension in R^N.

    Args:
        v1 (np.ndarray): N-D vector
        v2 (np.ndarray): N-D vector

    Returns:
        float: angle in radians
    """
    return np.arccos(np.dot(v1, v2) / (np.linalg.norm(v1) * np.linalg.norm(v2)))



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from typing import Callable

import numpy as np


class BaseStrain(ABC):
    """Base class for strain fields that act on Generators.

    Strain objects can be attached to generators, and transform the coordinates
    of the atoms post-generation of the supercell. Strain fields apply strain
    in crystal coordinate system by default.

    Attributes:
        origin (np.ndarray): origin with which respect coordinates are strained
        mode (str): "crystal" or "standard", for straining in crystal coordinates
                    or for straining coordinates with respect to standard R3
                    orthonromal basis and orientation, respectively
        bases (np.ndarray): 3x3 array representing generator basis vectors
    """

    def __init__(self):
        self.origin = np.array([0, 0, 0])

    @abstractmethod
    def apply_strain(self, points: np.ndarray) -> np.ndarray:
        """Apply strain to a collection of points.

        Args:
            points (np.ndarray): Nx3 array of points in space
        
        Returns:
            np.ndarray: Nx3 array of strained points in space
        """
        pass

    def scrape_params(self, obj: BaseGenerator):
        """Helper method to grab origin and bases from host generator.

        Args:
            obj (BaseGenerator): generator to grab parameters from
        """
        if self.mode == "crystal":
            self._bases = np.copy(obj.voxel.sbases)

        if self.origin == "generator":
            self.origin = np.copy(obj.origin)

    @property
    def origin(self):
        """Origin with respect to which strain is applied."""
        return self._origin

    @origin.setter
    def origin(self, val):
        assert (val.shape == (3,)), "Origin must have shape (3,)"
        self._origin = np.array(val)

    @property
    def mode(self):
        """Coordinate system for strain application, either 'crystal' or 'standard'."""
        return self._mode

    @mode.setter
    def mode(self, val):
        if val == "crystal" or "standard":
            self._mode = val
        else:
            raise ValueError("Mode must be either crystal or standard")

    @property
    def bases(self):
        """"Basis vectors of crystal coordinate system."""
        return self._bases


class HStrain(BaseStrain):
    """Strain class for applying homogeneous strain fields to generators.

    HStrain objects can be attached to generators, and transform the coordinates
    of the atoms post-generation of the supercell via simple strain tensor. 
    HStrain fields apply strain in crystal coordinate system by default.

    Attributes:
        matrix (np.ndarray): Matrix representing homogeneous strain tensor.
                            Can be set with 3 (x,y,z), 6 (Voigt notation), or
                            9 values (as list or 3x3 array).
    """

    def __init__(self, matrix=None, origin="generator", mode="crystal"):
        if not matrix is None:
            self.matrix = matrix
        else:
            # apply no strain
            self._matrix = np.eye(3)

        self.mode = mode

        if origin != "generator":
            self.origin = origin
        else:
            super.__init__()

        self._bases = None

    ##############
    # Properties #
    ##############

    @property
    def matrix(self):
        """Homogeneous strain tensor."""
        return self._matrix

    @matrix.setter
    def matrix(self, vals):
        vals = np.squeeze(np.array(vals))
        if vals.shape == (3,):
            self._matrix = np.eye(3) * vals
        elif vals.shape == (3, 3):
            self._matrix = vals
        elif vals.shape == (9,):
            self._matrix = np.reshape(vals, (3, 3))
        elif vals.shape == (6,):
            # voigt notation
            v = vals
            self._matrix = np.array([[v[0], v[5], v[4]], \
                                     [v[5], v[1], v[3]], \
                                     [v[4], v[3], v[2]]])
        else:
            raise ValueError("Input shape must be either 3,6, or 9 elements")

    ##############
    ### Methods ##
    ##############
    def apply_strain(self, points: np.ndarray) -> np.ndarray:

        # get points relative to origin
        sp = np.copy(points) - self.origin

        if self.mode == "crystal":
            # project onto crystal coordinates, strain, project back into real space
            sp = sp @ np.linalg.inv(self.bases) @ self.matrix @ self.bases
        else:
            # strain
            sp = sp @ self.matrix

        # shift back w.r.t. origin
        sp += self.origin

        return sp


class IStrain(BaseStrain):
    """Strain class for applying inhomogenous strain fields to generators.

    IStrain objects can be attached to generators, and transform the coordinates
    of the atoms post-generation of the supercell via arbitrary strain functions. 
    IStrain fields apply strain in crystal coordinate system by default.

    User must input a custom strain function; strain functions by default should
    accept only points as positional arguments and can take any kwargs.

    Attributes:
        fun_kwargs (dict): kwargs to pass to custom strain function
        strain_fun (Callable): strain function F: R3 -> R3 for 
                                np.arrays of shape (N,3)->(N,3)
    """

    def __init__(self, fun=None, origin="generator", mode="crystal", **kwargs):
        if not fun is None:
            self.strain_fun = fun
        else:
            # apply no strain
            self.strain_fun = lambda x: x

        self.mode = mode

        if origin != "generator":
            self.origin = origin
        else:
            super.__init__()

        self._bases = None
        self.fun_kwargs = kwargs

    ##############
    # Properties #
    ##############

    @property
    def fun_kwargs(self):
        """kwargs passed to custom strain function upon application of strain."""
        return self._fun_kwargs

    @fun_kwargs.setter
    def fun_kwargs(self, kwargs_dict: dict):
        assert (isinstance(
            kwargs_dict,
            dict)), "Must supply dictionary for arbirtrary extra kwargs"
        self._fun_kwargs = kwargs_dict

    @property
    def strain_fun(self):
        """Inhomogenous strain function to apply to coordinates."""
        return self._strain_fun

    @strain_fun.setter
    def strain_fun(self, fun: Callable[[np.ndarray], np.ndarray]):
        try:
            ref_arr = np.random.rand((100, 3))
            test_arr = fun(ref_arr, **self.fun_kwargs)
            assert (test_arr.shape == (100, 3))
        except AssertionError:
            raise ValueError(
                "Strain function must return numpy arrays with shape (N,3) for input arrays of shape (N,3)"
            )

        self._strain_fun = copy.deepcopy(fun)

    ##############
    ### Methods ##
    ##############
    def apply_strain(self, points: np.ndarray) -> np.ndarray:

        # get points relative to origin
        sp = np.copy(points) - self.origin

        if self.mode == "crystal":
            # project onto crystal coordinates
            sp = sp @ np.linalg.inv(self.bases)

            # strain
            sp = self.strain_fun(sp, basis=self.bases, **self.fun_kwargs)

            # project back into real space
            sp = sp @ self.bases
        else:
            # strain
            sp = self.strain_fun(sp, **self.fun_kwargs)

        # shift back w.r.t. origin
        sp += self.origin

        return sp



import matplotlib.pyplot as plt
import numpy as np
try:
    import ipywidgets as widgets
    from ipywidgets import (FloatSlider, Layout, fixed, interact, interact_manual,
                            interactive)
except:
    print("Could not import ipywidgets.")

from ..scene import Scene


def simple_view_with_rotate(data, azim=-60, elev=30):
    fig = plt.figure(figsize=(12, 3), constrained_layout=True)
    fig.set_constrained_layout_pads(w_pad=0.05,
                                    h_pad=0.05,
                                    hspace=0.1,
                                    wspace=0.2)
    ax_x = fig.add_subplot(1, 4, 3, projection='3d')
    ax_y = fig.add_subplot(1, 4, 2, projection='3d')
    ax_z = fig.add_subplot(1, 4, 1, projection='3d')
    ax_r = fig.add_subplot(1, 4, 4, projection='3d')

    ax_r.scatter(data[:, 0], data[:, 1], data[:, 2], s=30, alpha=1)
    ax_x.scatter(data[:, 0], data[:, 1], data[:, 2], s=30, alpha=1)
    ax_y.scatter(data[:, 0], data[:, 1], data[:, 2], s=30, alpha=1)
    ax_z.scatter(data[:, 0], data[:, 1], data[:, 2], s=30, alpha=1)

    ax_z.azim = -90.0
    ax_z.elev = 90.0

    ax_y.azim = -90.0
    ax_y.elev = 0.0

    ax_x.azim = 0.0
    ax_x.elev = 0.0

    ax_r.azim = azim
    ax_r.elev = elev
    ax_r.set_xticks([])
    ax_r.set_yticks([])
    ax_r.set_zticks([])

    ax_x.set_xticks([])
    ax_y.set_yticks([])
    ax_z.set_zticks([])

    ax_x.grid(True)
    ax_y.grid(True)
    ax_z.grid(True)

    ax_x.set_ylabel("Y")
    ax_x.set_zlabel("Z")

    ax_y.set_xlabel("X")
    ax_y.set_zlabel("Z")

    ax_z.set_xlabel("X")
    ax_z.set_ylabel("Y")

    ax_x.set_title("YZ projection")
    ax_y.set_title("XZ projection")
    ax_z.set_title("XY projection")
    ax_r.set_title("Rotatable view")

    plt.draw()


def simple_scene_view(scene, azim=-60, elev=30):
    if not isinstance(scene, Scene):
        raise TypeError("Supplied scene not a Scene() object")

    data = scene.all_atoms
    species = scene.all_species
    bbox = scene.bounds

    fig = plt.figure(figsize=(12, 3), constrained_layout=True)
    fig.set_constrained_layout_pads(w_pad=0.05,
                                    h_pad=0.05,
                                    hspace=0.1,
                                    wspace=0.2)

    ax_x = fig.add_subplot(1, 4, 3, projection='3d')
    ax_y = fig.add_subplot(1, 4, 2, projection='3d')
    ax_z = fig.add_subplot(1, 4, 1, projection='3d')
    ax_r = fig.add_subplot(1, 4, 4, projection='3d')

    for s in np.unique(species):
        mask = species == s
        ax_r.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_x.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_y.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_z.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)

    ax_z.azim = -90.0
    ax_z.elev = 90.0

    ax_y.azim = -90.0
    ax_y.elev = 0.0

    ax_x.azim = 0.0
    ax_x.elev = 0.0

    ax_r.azim = azim
    ax_r.elev = elev

    ax_r.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_r.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_r.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_x.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_x.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_x.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_y.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_y.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_y.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_z.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_z.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_z.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_r.set_xticks([])
    ax_r.set_yticks([])
    ax_r.set_zticks([])

    ax_x.set_xticks([])
    ax_y.set_yticks([])
    ax_z.set_zticks([])

    ax_x.grid(True)
    ax_y.grid(True)
    ax_z.grid(True)

    ax_x.set_ylabel("Y")
    ax_x.set_zlabel("Z")

    ax_y.set_xlabel("X")
    ax_y.set_zlabel("Z")

    ax_z.set_xlabel("X")
    ax_z.set_ylabel("Y")

    ax_x.set_title("YZ projection")
    ax_y.set_title("XZ projection")
    ax_z.set_title("XY projection")
    ax_r.set_title("Rotatable view")

    plt.draw()


def simple_scene_view_with_rotate(scene, azim=-60, elev=30):
    if not isinstance(scene, Scene):
        raise TypeError("Supplied scene not a Scene() object")

    data = scene.all_atoms
    species = scene.all_species
    bbox = scene.bounds

    fig = plt.figure(figsize=(12, 3), constrained_layout=True)
    fig.set_constrained_layout_pads(w_pad=0.05,
                                    h_pad=0.05,
                                    hspace=0.1,
                                    wspace=0.2)

    ax_x = fig.add_subplot(1, 4, 3, projection='3d')
    ax_y = fig.add_subplot(1, 4, 2, projection='3d')
    ax_z = fig.add_subplot(1, 4, 1, projection='3d')
    ax_r = fig.add_subplot(1, 4, 4, projection='3d')

    for s in np.unique(species):
        mask = species == s
        ax_r.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_x.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_y.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)
        ax_z.scatter(data[mask, 0], data[mask, 1], data[mask, 2], s=30, alpha=1)

    ax_z.azim = -90.0
    ax_z.elev = 90.0

    ax_y.azim = -90.0
    ax_y.elev = 0.0

    ax_x.azim = 0.0
    ax_x.elev = 0.0

    ax_r.azim = azim
    ax_r.elev = elev

    ax_r.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_r.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_r.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_x.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_x.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_x.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_y.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_y.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_y.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_z.set_xlim([bbox[0, 0], bbox[1, 0]])
    ax_z.set_ylim([bbox[0, 1], bbox[1, 1]])
    ax_z.set_zlim([bbox[0, 2], bbox[1, 2]])

    ax_r.set_xticks([])
    ax_r.set_yticks([])
    ax_r.set_zticks([])

    ax_x.set_xticks([])
    ax_y.set_yticks([])
    ax_z.set_zticks([])

    ax_x.grid(True)
    ax_y.grid(True)
    ax_z.grid(True)

    ax_x.set_ylabel("Y")
    ax_x.set_zlabel("Z")

    ax_y.set_xlabel("X")
    ax_y.set_zlabel("Z")

    ax_z.set_xlabel("X")
    ax_z.set_ylabel("Y")

    ax_x.set_title("YZ projection")
    ax_y.set_title("XZ projection")
    ax_z.set_title("XY projection")
    ax_r.set_title("Rotatable view")

    plt.draw()


def simple_view_widget(data):
    azim_widget = FloatSlider(value=-60,
                              min=-180,
                              max=180,
                              step=0.1,
                              description='Azimuth',
                              readout_format='.1f',
                              style={'description_width': '150px'},
                              layout=Layout(width='400px', height='30px'))

    elev_widget = FloatSlider(value=30,
                              min=-90,
                              max=90,
                              step=0.1,
                              description='Elevation',
                              readout_format='.1f',
                              style={'description_width': '150px'},
                              layout=Layout(width='400px', height='30px'))

    return interact(simple_view_with_rotate,
                    data=fixed(data),
                    azim=azim_widget,
                    elev=elev_widget)


def simple_scene_widget(scene):
    azim_widget = FloatSlider(value=-60,
                              min=-180,
                              max=180,
                              step=0.1,
                              description='Azimuth',
                              readout_format='.1f',
                              style={'description_width': '150px'},
                              layout=Layout(width='400px', height='30px'))

    elev_widget = FloatSlider(value=30,
                              min=-90,
                              max=90,
                              step=0.1,
                              description='Elevation',
                              readout_format='.1f',
                              style={'description_width': '150px'},
                              layout=Layout(width='400px', height='30px'))

    return interact(simple_scene_view_with_rotate,
                    scene=fixed(scene),
                    azim=azim_widget,
                    elev=elev_widget)



from typing import List

import numpy as np
from ase import Atoms
from ase.io import write as ase_write
from ase.symbols import Symbols

from ..volume import BaseVolume, makeRectPrism


class Scene():
    """Scene classes manage multiple objects interacting in space with cell boundaries.

    Attributes:
        bounds (np.ndarray): 2x3 array defining rectangular bounds of scene.
        objects (List[BaseVolume]): List of all objects currently in scene.
        all_atoms (np.ndarray): Coordinates of all atoms in scene after precedence checks.
        all_species (np.ndarray): Atomic numbers of all atoms in scene after precedence checks.
        ase_atoms (Atoms): Collection of atoms in scene as ASE Atoms object.
    """

    def __init__(self, bounds=None, objects=None):
        self._bounds = None
        self._objects = None
        self._checks = None

        if not (objects is None):
            if (hasattr(objects, "__iter__")):
                for ob in objects:
                    self.add_object(ob)
            else:
                self.add_object(objects)

        if bounds is None:
            #default bounding box is 10 angstrom cube
            bbox = makeRectPrism(10, 10, 10)
            self.bounds = np.vstack(
                [np.min(bbox, axis=0),
                 np.max(bbox, axis=0)])
        else:
            self.bounds = bounds

    @property
    def bounds(self):
        """Current boundaries of nanoscale scene."""
        return self._bounds

    @bounds.setter
    def bounds(self, bounds: np.ndarray):
        bounds = np.array(bounds)
        assert (bounds.shape == (2, 3))
        self._bounds = bounds

    @property
    def objects(self):
        """List of objects in current scene."""
        return self._objects

    def add_object(self, ob: BaseVolume):
        """Add an object to the scene.
        
        Args:
            ob (BaseVolume): object to add to scene.
        """
        #for now, only volumes are objects
        if isinstance(ob, BaseVolume):
            if self._objects is None:
                self._objects = [ob]
            else:
                self._objects.append(ob)

    @property
    def _checks(self):
        """List of logical arrays indicating inclusion of atoms in scene from each object."""
        return self._checks

    @property
    def all_atoms(self):
        """Positions of all atoms currently in the scene after evaluating conflict resolution."""
        return np.vstack(
            [ob.atoms[self._checks[i], :] for i, ob in enumerate(self.objects)])

    @property
    def all_species(self):
        """Atomic numbers of all atoms currently in the scene after evaluating conflict resolution."""
        return np.hstack(
            [ob.species[self._checks[i]] for i, ob in enumerate(self.objects)])

    @property
    def ase_atoms(self):
        """Collection of atoms in scene as ASE Atoms object."""
        cell_dims = self.bounds[1, :] - self.bounds[0, :]
        return Atoms(symbols=self.all_species,
                     positions=self.all_atoms,
                     cell=cell_dims)

    def species_from_object(self, idx: int):
        """Grab all the atoms from contributing object at idx.
        
        Returns:
            Numpy array of all positions of atoms contributed by object at idx.
        """
        return self.objects[idx].atoms[self._checks[idx], :]

    def _get_priorities(self):
        """Grab priority levels of all objects in Scene to determine precedence relationship.

        Returns:
            List of relative priority levels and offsets. Relative priority levels
            and offsets are used to determine which objects whill be checked
            for the inclusion of atoms in the scene of the atoms contributed by
            another object.
        
        """
        # get all priority levels active first
        self.objects.sort(key=lambda ob: ob.priority)
        plevels = np.array([x.priority for x in self.objects])

        # get unique levels and create relative priority array
        __, idx = np.unique(plevels, return_index=True)
        rel_plevels = np.zeros(len(self.objects)).astype(int)
        for i in idx[1:]:
            rel_plevels[i:] += 1

        offsets = np.append(idx, len(self.objects))

        return rel_plevels, offsets

    def populate(self):
        """Populate the scene with atoms according to Volumes and priority levels.

        First, every object populates atoms against its own boundaries.
        Then, gather the list of priorities from all the objects.
        For each object, generate a True array of length ob.atoms. 
        For each object in the same priority level or lower, perform interiority 
        check and repeatedly perform logical_and to see if atoms belong in scene.

        - Lower priority numbers supercede objects with high priority numbers.
        - Objects on the same priority level will not supply atoms to the scene in their volume intersections.
        """
        for ob in self.objects:
            ob.populate_atoms()

        rel_plevels, offsets = self._get_priorities()

        self._checks = []

        for i, ob in enumerate(self.objects):
            check = np.ones(ob.atoms.shape[0]).astype(bool)
            eidx = offsets[rel_plevels[i] + 1]

            for j in range(eidx):
                if (i != j):
                    check_against = np.logical_not(
                        self.objects[j].checkIfInterior(ob.atoms))
                    check = np.logical_and(check, check_against)

            self._checks.append(check)

    def to_file(self, fname, **kwargs):
        """Write atomic scene to an output file, using ASE write utilities.

        If format="prismatic", will default to Debye-Waller factors of 0.1 RMS 
        displacement in squared angstroms, unless dictionary of debye-waller factors
        is otherwise supplied.
        
        Args:
            fname (str): output file name.
            **kwargs: any key word arguments otherwise accepted by ASE write.
        """
        if "format" in kwargs.keys():
            if kwargs["format"] == "prismatic":
                dwf = set(self.all_species)
                dw_default = (0.1**2.0) * 8 * np.pi**2.0
                dwf = {str(Symbols([x])): dw_default for x in dwf}
                ase_write(filename=fname,
                          images=self.ase_atoms,
                          debye_waller_factors=dwf,
                          **kwargs)
        else:
            ase_write(filename=fname, images=self.ase_atoms, **kwargs)



from typing import List

import numpy as np

############################
######### Utilities ########
############################


def get_p_dist(coords, new_coord, dims):
    """Calculate squared periodic distance between a coordinate and all other coordinates in a set."""

    dist_0 = np.abs(coords - new_coord)
    dist_1 = dims - dist_0
    dist_1[:, 2] = dist_0[:, 2]

    p_dist = np.sum(np.min(np.dstack([dist_0, dist_1]), axis=2)**2.0, axis=1)
    return p_dist


def get_tuples(tx, ty, tz):

    tlist = [(x, y, z) for x in tx for y in ty for z in tz]

    return tlist


############################################
######## Periodic Uniform Algorithm ########
############################################


def get_voxels(min_dist, dims):
    """Get voxel and neighbor list."""
    num_blocks = np.ceil(dims / (min_dist)).astype(int)
    voxels = []
    neighbors = []
    for i in range(num_blocks[0]):
        voxels.append([])
        neighbors.append([])
        for j in range(num_blocks[1]):
            voxels[-1].append([])
            neighbors[-1].append([])
            for k in range(num_blocks[2]):
                voxels[-1][-1].append([])

                tmp_x = [i, (i + 1) % num_blocks[0], (i - 1) % num_blocks[0]]
                tmp_y = [j, (j + 1) % num_blocks[1], (j - 1) % num_blocks[1]]
                tmp_z = [k]
                if (k > 0):
                    tmp_z.append(k - 1)
                if (k < num_blocks[2] - 1):
                    tmp_z.append(k + 1)

                tlist = get_tuples(tmp_x, tmp_y, tmp_z)

                neighbors[-1][-1].append(tlist)

    return voxels, neighbors


def gen_p_substrate(dims: List[float],
                    min_dist: float = 1.4,
                    density=.1103075,
                    print_progress=True):
    """Generate a uniformly random distributed collection of atoms with PBC.

    Given the size of a rectangular prism, a minimum bond distance, and a target
    density, generate a uniformly random collection of atoms obeying periodic
    boundary conditions in X and Y. Dimensions, minimum distance, and density
    should all be in units of angstroms but can be input in any consistent unit 
    scheme. Default values are for amorphous carbon.

    Generation algorithm loosely follows
      1. Get total number  of atoms N to generate.
      2. While substrate contains < N atoms
        a. Generate uniformly random coordinate
        b. Check distance against nearest neighbor atoms to for violation
          of bond distance
        c. If not too close to other atoms, add to substrate; else, regenerate

    Generation utilizes voxel grid in 3D space for linear scaling of distance
    calculations and therefore generation time should loosely scale linearly
    with the volume of the substrate.

    Args:
        dims (List[float]): Size of rectangular prism substrate in [x,y,z]
        min_dist (float): Minimum seperation between atoms.
        density (float): Density of substrate.

    Returns:
        np.ndarray: coordinates of atoms in periodic substrate
    """

    # get number of carbon atoms to generate
    dims = np.array(dims)
    min_dist_2 = min_dist**2.0
    dim_x = dims[0]
    dim_y = dims[1]
    dim_z = dims[2]
    sub_vol = dim_x * dim_y * dim_z

    num_c = np.round(sub_vol * density).astype(int)
    coords = np.zeros((num_c, 3))
    dims = np.array([dim_x, dim_y, dim_z])

    if print_progress:
        print("Getting neighbors")

    # get voxel grid and list of local voxel neighbors
    voxels, neighbors = get_voxels(min_dist, dims)

    coords[0, :] = np.random.random_sample((1, 3)) * dims
    if print_progress:
        print("Starting particle loop for %i particles" % num_c)

    for i in range(1, num_c):
        if print_progress and (not (i % (num_c // 5))):
            print("On %i of %i" % (i, num_c))
        new_coord = np.random.random_sample((1, 3)) * dims

        block = np.floor(new_coord / (min_dist)).astype(int)
        block = block[0]
        tlist = neighbors[block[0]][block[1]][block[2]]
        parts = []
        for t in tlist:
            parts.extend(voxels[t[0]][t[1]][t[2]])

        p_dist = get_p_dist(coords[parts, :], new_coord, dims)
        while ((len(p_dist) > 0) and (np.min(p_dist) < min_dist_2)):
            new_coord = np.random.random_sample((1, 3)) * dims
            block = np.floor(new_coord / (min_dist)).astype(int)
            block = block[0]
            tlist = neighbors[block[0]][block[1]][block[2]]
            parts = []
            for t in tlist:
                parts.extend(voxels[t[0]][t[1]][t[2]])

            p_dist = get_p_dist(coords[parts, :], new_coord, dims)

        voxels[block[0]][block[1]][block[2]].append(i)
        coords[i, :] = new_coord

    return coords



from abc import ABC, abstractmethod
from typing import Dict, List, Tuple

import numpy as np
from ase import Atoms
from wulffpack import Decahedron, Icosahedron, SingleCrystal, Winterbottom
from wulffpack.core import BaseParticle

from ..generator.generator import Generator
from ..transform.strain import HStrain, IStrain
from ..transform.transform import MatrixTransform, Rotation, rot_v
from ..volume.algebraic import Cylinder, Plane, Sphere
from ..volume.volume import MultiVolume, Volume

"""
Submodule provides convenience constructors for returning Wulff constructions
as calculated and implemented by Wulffpack in their respective forms as 
Construction Zone volumes and generators
"""


class WulffBase(ABC):
    """Base class for Wulff constructions, as adapted from WulffPack.
    
    Wulff constructions are energetic constructions of single crystals. From a 
    list of surfaces and corresponding surface energies, the equilibirium shape
    minimizing the total surface energy for a given volume is determined. The
    equilbirium shape is looseley the inner surface of the intersection of the 
    supplied surfaces, where the supplied surfaces are translated from the origin
    along their normal directions by their relative surface energies.

    Attributes:
        generator (Generator): Crystalline generator.
        p_atoms (Atoms): Primitive structure of generator, as ASE Atoms object.
        natoms (int): Approximate size of particle in number of atoms.
        surface_energies (Dict[Tuple, float]): dictionary of surfaces and corresponding
                                            surface energies, where keys are
                                            tuples of miller indices and values
                                            are surfaces energies as floats
    """

    @abstractmethod
    def __init__(self):
        pass

    ####################
    #### Properties ####
    ####################

    @property
    def generator(self):
        """Generator object representing crystal system."""
        return self._generator

    @generator.setter
    def generator(self, val):
        assert (isinstance(val, Generator)
               ), "Supplied generator must be crystalline Generator class"
        self._generator = val
        p_struct = val.structure.get_primitive_structure()
        self._p_atoms = Atoms(positions=p_struct.cart_coords)

    @property
    def p_atoms(self):
        """Primitive structure of generator as ASE Atoms object."""
        return self._p_atoms

    @property
    def natoms(self):
        """Size of particle in approximate number of atoms."""
        return self._natoms

    @natoms.setter
    def natoms(self, val):
        try:
            self._natoms = int(val)
        except:
            raise TypeError("natoms must be convertible to integer")

    @property
    def surface_energies(self):
        """Dictionary of surface: surface energy pairs."""
        return self._surface_energies

    @surface_energies.setter
    def surface_energies(self, vals):
        for key, val in vals.items():
            assert(isinstance(key, tuple) and isinstance(val, float)), \
                "surface energies must be passed as tuple pairs of miller indices and floats"

        self._surface_energies = vals

    ####################
    ###### Methods #####
    ####################

    @staticmethod
    def planes_from_wulff(wulff):
        """Return facets of a full Wulff construction as list of Planes.

        Args:
            wulff (BaseParticle): WulffPack particle.

        Returns:
            List of Planes representing facets of Wulff constructions.
        """
        return [
            Plane(f.normal, f.normal * f.distance_from_origin)
            for f in wulff._yield_facets()
        ]

    @staticmethod
    def planes_from_form(wulff_form):
        """Return facets of a full Wulff construction as list of Planes.

        Args:
            wulff (BaseParticle): WulffPack particle.

        Returns:
            List of Planes representing facets of Wulff constructions.
        """
        return [
            Plane(f.normal, f.normal * f.distance_from_origin)
            for f in wulff_form.facets
        ]

    @abstractmethod
    def get_construction(self):
        """For a given crystal and set of surface energies, return a Wulff construction.

        Returns:
            Volume or MultiVolume with planes and generator with object prescribed 
            by given Wulff Construction.
        """
        pass


class WulffSingle(WulffBase):

    def __init__(self,
                 generator: Generator,
                 surface_energies: dict,
                 natoms: int = 1000):
        self.generator = generator
        self.natoms = natoms
        self.surface_energies = surface_energies

    ####################
    #### Properties ####
    ####################

    ####################
    ###### Methods #####
    ####################

    def get_construction(self):
        wulff = SingleCrystal(self.surface_energies, self.p_atoms, self.natoms)
        planes = self.planes_from_wulff(wulff)
        return Volume(alg_objects=planes, generator=self.generator)

    def winterbottom(self, interface: Tuple[int], interface_energy: float):
        """Construct Winterbottom construction for single crystals.
        
        A Winterbottom construction is analagous to a Wulff construction, in that
        it is the shape for a single crystal particle that minimizes the surface
        energy when the particle is in contact with another surface.

        Args:
            interface (Tuple[int]): Miller indices of interfacial surface
            interface_energy: Relative surface energy of interface. Must be 
                            less than or equal to half of the smallest surface
                            energy for the other surfaces.

        Returns:
            Volume with shape described by Winterbottom construction for given 
            surfce energies.
        """
        wulff = Winterbottom(self.surface_energies, interface, interface_energy,
                             self.p_atoms, self.natoms)
        planes = self.planes_from_wulff(wulff)

        return Volume(alg_objects=planes, generator=self.generator)

    @classmethod
    def cube(cls, generator: Generator, natoms: int = 1000):
        """Convenience constructor for cubic nanoparticles with Wulff construction.
        
        Args:
            generator (Generator): crystalline system for Wulff construction.
            natoms (int): approximate size of particle in atoms.

        Returns:
            Wulff construction for a cubic nanoparticle.
        """
        energies = {(1, 0, 0): 1.0}
        wulff = cls(generator, natoms)
        return wulff.get_construction(energies)

    @classmethod
    def cuboctohedron(cls, generator: Generator, natoms: int = 1000):
        """Convenience constructor for cuboctohedra; nanoparticles with Wulff construction.

        Args:
            generator (Generator): crystalline system for Wulff construction.
            natoms (int): approximate size of particle in atoms.

        Returns:
            Wulff construction for a cuboctohedral nanoparticle.
        """
        energies = {(1, 0, 0): 1.0, (1, 1, 1): 1.15}
        wulff = cls(generator, natoms)
        return wulff.get_construction(energies)

    @classmethod
    def octohedron(cls, generator: Generator, natoms: int = 1000):
        """Convenience constructor for octohedral nanoparticles with Wulff construction.
        
        Args:
            generator (Generator): crystalline system for Wulff construction.
            natoms (int): approximate size of particle in atoms.

        Returns:
            Wulff construction for a octohedral nanoparticle.
        """
        energies = {(1, 1, 1): 1.0}
        wulff = cls(generator, natoms)
        return wulff.get_construction(energies)

    @classmethod
    def truncated_octohedron(cls, generator: Generator, natoms: int = 1000):
        """Convenience constructor for truncted octohedral nanoparticles with Wulff construction.
        
        Args:
            generator (Generator): crystalline system for Wulff construction.
            natoms (int): approximate size of particle in atoms.

        Returns:
            Wulff construction for a truncated octohedral nanoparticle.
        """
        energies = {(1, 0, 0): 1.1, (1, 1, 0): 1.15, (1, 1, 1): 1.0}
        wulff = cls(generator, natoms)
        return wulff.get_construction(energies)


"""
WulffPack's dechaedron and icosahedron class is not as clean to grab.
Generally, they form the nanoparticle twice, in independent steps.
For facet visualization, they generate the first grain's surfaces as facets,
then rotate the group of facets about the central axes. 

Since the number of rotations applied is known, and the grains are stored in order,
we can back out which facets (and which rotations) are applied to each grain.
This will yield us all (most) of the subvolumes we need for the particle,
minus the internal zone of the rotation axis.

For atomic generation, they work again with a single grain. They first remove
atoms on the twin boundaries, and store atoms within the grain/inside the fivefold axis separately.
That is, twin_indices contains twin boundaries, which contains five_old axis atoms, too.
They then strain the in grain atoms out, make rotated duplicates, and return the atoms only.

The five fold axis is just a stack of atoms along the center. We should be able to replicate this in CZ with a 
cylinder object, placed at the origin and with diameter equal to the atom spacing.
####################################################

Strategy for replication of Decahedron:
1) Get list of facets.
2) Prune list into the 5 separate grains, create subvolumes
4) Create strained generator, rotate it to create grain specific generators, add to subvolumes
5) Create subvolume for 5 fold axis based on Cylinder, add unstrained generator
6) Return MultiVolume object with mutually exclusive priorities

"""


class WulffDecahedron(WulffBase):
    """Prefab routine for constructing strained Decahedra for FCC systems.
    
    Decahedral nanoparticles cannot be created in a defect-free, strain-free manner;
    however, relatively simple, physically close models can be generated through
    simple strain and twin defect models with Wulff constructions defining 
    tetrahedral units. Construction algorithm is an adapatation of that found
    in WulffPack. 
    
    See L. D. Marks,  Modified Wulff constructions for twinned particles,
    J. Cryst. Growth 61, 556 (1983), doi: 10.1016/0022-0248(83)90184-7 for 
    background.

    Attributes:
        generator (Generator): Crystalline generator.
        p_atoms (Atoms): Primitive structure of generator, as ASE Atoms object.
        natoms (int): Approximate size of particle in number of atoms.
        surface_energies (Dict[Tuple, float]): dictionary of surfaces and corresponding
                                            surface energies, where keys are
                                            tuples of miller indices and values
                                            are surfaces energies as floats.
        twin_energy (float): surface energy of twin boundaries in system.
    """

    def __init__(self,
                 generator: Generator,
                 surface_energies: dict,
                 twin_energy: float,
                 natoms: int = 1000):
        self.generator = generator
        self.natoms = natoms
        self.surface_energies = surface_energies
        self.twin_energy = twin_energy

    @property
    def twin_energy(self):
        """Surface energy of twin boundary defects."""
        return self._twin_energy

    @twin_energy.setter
    def twin_energy(self, val):
        assert (isinstance(val, float)), "twin energy must be float"
        self._twin_energy = val

    def get_construction(self):
        wulff = Decahedron(self.surface_energies, self.twin_energy,
                           self.p_atoms, self.natoms)
        sf = wulff._get_dechedral_scale_factor()
        planes_lists = [[], [], [], [], []]
        for form in wulff.forms:
            planes = self.planes_from_form(form)
            len_p = len(planes)
            for i, p in enumerate(planes):
                # there will always be a multiple 5 planes in planes
                idx = i // (len_p // 5)
                planes_lists[idx].append(p)

        vols = []
        for plist in planes_lists:
            vols.append(Volume(alg_objects=plist))

        # create strained and rotated generators
        strain_field = HStrain(matrix=[sf, 1, 1, 0, 0, 0])

        for i in range(0, 5):
            rot_mat = rot_v([0, 0, 1], i * 2 * np.pi / 5)
            rot = Rotation(rot_mat)
            gen = self.generator.from_generator(transformation=[rot])
            gen.strain_field = strain_field
            vols[i].add_generator(gen)

        z_extremes = np.array([
            np.max(np.vstack(f.vertices)[:, 2]) for f in wulff._yield_facets()
        ])
        ff_axis_b = Plane(normal=[0, 0, -1], point=[0, 0, np.min(z_extremes)])
        ff_axis_t = Plane(normal=[0, 0, 1], point=[0, 0, np.max(z_extremes)])
        ff_axis_cyl = Cylinder(radius=1e-3)
        vols.append(
            Volume(alg_objects=[ff_axis_b, ff_axis_t, ff_axis_cyl],
                   generator=self.generator))

        return MultiVolume(volumes=vols)


class WulffIcosahedron(WulffDecahedron):
    """Prefab routine for constructing strained Decahedra for FCC systems.
    
    Icosahedral nanoparticles cannot be created in a defect-free, strain-free manner;
    however, relatively simple, physically close models can be generated through
    simple strain and twin defect models with Wulff constructions defining 
    tetrahedral units. Construction algorithm is an adapatation of that found
    in WulffPack. 
    
    See L. D. Marks,  Modified Wulff constructions for twinned particles,
    J. Cryst. Growth 61, 556 (1983), doi: 10.1016/0022-0248(83)90184-7 for 
    background.

    Attributes:
        generator (Generator): Crystalline generator.
        p_atoms (Atoms): Primitive structure of generator, as ASE Atoms object.
        natoms (int): Approximate size of particle in number of atoms.
        surface_energies (Dict[Tuple, float]): dictionary of surfaces and corresponding
                                            surface energies, where keys are
                                            tuples of miller indices and values
                                            are surfaces energies as floats
        twin_energy (float): surface energy of twin boundaries in system.
    """

    @property
    def icosahedral_scale_factor(self):
        k = (5 + np.sqrt(5)) / 8
        return np.sqrt(2 / (3 * k - 1))

    @staticmethod
    def _strain_from_111(points, basis=np.eye(3).astype(float), sf=1):
        """Inhomogenous strain function to strain atoms away from [111] axis.
        
        Args:
            points (np.ndarray): Nx3 array of points
            basis (np.ndarray): 3x3 array representing crystal basis vectors
            sf (float): scale factor for icosahedral particle.

        Returns:
            np.ndarray: Nx3 set of strained points
        """
        vec_111 = basis @ np.array([[1.0, 1.0, 1.0]]).T
        vec_111 /= np.linalg.norm(vec_111)

        disp_vec = points - np.dot(points, vec_111) * vec_111.T
        return points + (sf - 1.0) * disp_vec

    def get_construction(self):
        wulff = Icosahedron(self.surface_energies, self.twin_energy,
                            self.p_atoms, self.natoms)

        sf = wulff._get_icosahedral_scale_factor()
        planes_lists = [[] for x in range(20)]

        for form in wulff.forms:
            planes = self.planes_from_form(form)
            len_p = len(planes)
            for i, p in enumerate(planes):
                # there will always be a multiple 20 planes in planes
                idx = i // (len_p // 20)
                planes_lists[idx].append(p)

        vols = []
        for plist in planes_lists:
            vols.append(Volume(alg_objects=plist))

        # get strained and rotated generators for the 20 tetrahedral grains
        strain_field = IStrain(fun=self._strain_from_111(),
                               sf=self.icosahedral_scale_factor())

        sym_mats = [MatrixTransform(np.eye(3))] + \
                    [MatrixTransform(m) for m in wulff._get_all_symmetry_operations()]

        for i, m in enumerate(sym_mats):
            gen = self.generator.from_generator(transformation=m)
            gen.strain_field = strain_field

            # setting an ordered priority might ensure that the fivefold axes, twin boundaries aren't duplicated
            # TODO: test this against wulffpack implementation
            vols[i].add_generator(gen)
            vols[i].priority = i

        central_sphere = Volume(alg_objects=Sphere(radius=1.0),
                                generator=self.generator,
                                priority=-1)
        vols.append(central_sphere)

        return MultiVolume(volumes=vols)



from collections.abc import Callable
from typing import List, Tuple

import numpy as np


def get_voxel_grid(dim: Tuple[int],
                   px: bool = True,
                   py: bool = True,
                   pz: bool = True):
    """Generate list of voxel neighbors for 3D voxel grid with periodic boundary conditions.

    Utility function which returns a representation of a connected 3D voxel grid
    for arbitrary periodic boundary conditions. Voxels are ordered on a 1D list
    by X, then Y, then Z. For example, a 2x2x2 voxel grid will be ordered as
    [(0,0,0), (1,0,0), (0,1,0), (1,1,0), (0,0,1), (1,0,1), (1,1,1)]. An ordered
    list is returned which contains, for each voxel i, a list of all of its 
    neighbors on the 3D grid, as ordered in the 1D indexing scheme.

    For fully periodic boundary conditions, each voxel i will have 27 neighbors, 
    including the voxel index itself.


    Args:
        dim (Tuple[int]): size of grid in x, y, and z
        px (bool): periodicity in x
        py (bool): periodicity in y
        pz (bool): periodicity in z

    Returns:
        List[List[int]]
    """
    # get relative coordinates as 3D grid
    nn = [x for x in range(27)]
    nn_x = [(x % 3) - 1 for x in nn]
    nn_y = [((x // 3) % 3) - 1 for x in nn]
    nn_z = [(x // 9) - 1 for x in nn]

    N = np.prod(dim)
    neighbors = np.ones((N, 27)) * np.arange(N)[:, None]

    # get relative indices as 1D list
    shifts = (1, dim[0], dim[0] * dim[1])
    for i, t in enumerate(zip(nn_x, nn_y, nn_z)):
        neighbors[:, i] += np.dot(t, shifts)

    # correct list for total size of grid
    neighbors = neighbors % N

    # if fully periodic, no further corrections needed
    if px and py and pz:
        return neighbors

    # get full list of indices as array
    idx = np.array([x for x in range(N)]).astype(int)

    # get logical arrays for voxels on faces of grid
    xi_face = (idx % dim[0] == 0)[:, None]
    xf_face = (idx % dim[0] == dim[0] - 1)[:, None]

    yi_face = ((idx // dim[0]) % dim[1] == 0)[:, None]
    yf_face = ((idx // dim[0]) % dim[1] == dim[1] - 1)[:, None]

    zi_face = (idx // (dim[0] * dim[1]) == 0)[:, None]
    zf_face = (idx // (dim[0] * dim[1]) == dim[2] - 1)[:, None]

    # get local neighbors on faces as arrays so that we can use matrix
    # multiplication for logical indexing
    nn_xi = np.array([(x % 3) == 0 for x in nn])[None, :]
    nn_xf = np.array([(x % 3) == 2 for x in nn])[None, :]

    nn_yi = np.array([((x // 3) % 3) == 0 for x in nn])[None, :]
    nn_yf = np.array([((x // 3) % 3) == 2 for x in nn])[None, :]

    nn_zi = np.array([(x // 9) == 0 for x in nn])[None, :]
    nn_zf = np.array([(x // 9) == 2 for x in nn])[None, :]

    # change neighbor indices to nans if boundary is not periodic
    fx = not px
    fy = not py
    fz = not pz

    # correct just x faces
    if fx:
        neighbors[xi_face @ nn_xi] = np.nan
        neighbors[xf_face @ nn_xf] = np.nan

    # correct just y faces
    if fy:
        neighbors[yi_face @ nn_yi] = np.nan
        neighbors[yf_face @ nn_yf] = np.nan

    # correct just z faces
    if fz:
        neighbors[zi_face @ nn_zi] = np.nan
        neighbors[zf_face @ nn_zf] = np.nan

    # correct x/y edges
    if fx and fy:
        for xc, nn_xc in zip([xi_face, xf_face], [nn_xi, nn_xf]):
            for yc, nn_yc in zip([yi_face, yf_face], [nn_yi, nn_yf]):
                idx_check = np.logical_and(xc, yc)
                nn_check = np.logical_and(nn_xc, nn_yc)
                neighbors[idx_check @ nn_check] = np.nan

    # correct x/z edges
    if fx and fz:
        for xc, nn_xc in zip([xi_face, xf_face], [nn_xi, nn_xf]):
            for zc, nn_zc in zip([zi_face, zf_face], [nn_zi, nn_zf]):
                idx_check = np.logical_and(xc, zc)
                nn_check = np.logical_and(nn_xc, nn_zc)
                neighbors[idx_check @ nn_check] = np.nan

    # correct y/z edges
    if fy and fz:
        for yc, nn_yc in zip([yi_face, yf_face], [nn_yi, nn_yf]):
            for zc, nn_zc in zip([zi_face, zf_face], [nn_zi, nn_zf]):
                idx_check = np.logical_and(yc, zc)
                nn_check = np.logical_and(nn_yc, nn_zc)
                neighbors[idx_check @ nn_check] = np.nan

    # correct x/y/z corners
    if fx and fy and fz:
        for xc, nn_xc in zip([xi_face, xf_face], [nn_xi, nn_xf]):
            for yc, nn_yc in zip([yi_face, yf_face], [nn_yi, nn_yf]):
                for zc, nn_zc in zip([zi_face, zf_face], [nn_zi, nn_zf]):
                    idx_check = np.logical_and(np.logical_and(yc, zc), xc)
                    nn_check = np.logical_and(np.logical_and(nn_yc, nn_zc),
                                              nn_xc)
                    neighbors[idx_check @ nn_check] = np.nan

    # used masked arrays to get compact lists of neighboring voxels
    mask = np.isnan(neighbors)
    neighbors_ma = np.ma.masked_array(neighbors, mask=mask).astype(int)
    neighbor_lists = [np.ma.compressed(x) for x in neighbors_ma]

    return neighbor_lists


def get_sdist_fun(dims=None, px=False, py=False, pz=False):
    """Return a squared distance function in 3D space for any PBC.

    Args:
        dims (Array[float]): length 3 array, list, or tuple for periodic
                            dimension sizes (x,y,z); must be passed if using
                            periodic boundaries
        px (bool): whether or not periodic in x
        py (bool): whether or not periodic in y
        pz (bool): whether or not periodic in z

    Returns:
        Callable[float] : squared distance function obeying periodic boundaries
        
    """
    # return standard distance function if no periodic bonadries
    if not np.any((px, py, pz)):

        def sdist(A, B):
            return np.sum((A - B)**2.0, axis=1)

    # grab relevant dimensions
    cols = [x for x, y in zip([0, 1, 2], [px, py, pz]) if y]
    sdims = np.array(dims)[cols]
    sdims = sdims[None, :, None]

    def sdist(A, B):
        dist_0 = np.abs(A - B)
        dist_1 = sdims - dist_0[:, cols, :]
        dist_0[:, cols, :] = np.min(np.stack([dist_0[:, cols, :], dist_1],
                                             axis=-1),
                                    axis=-1)

        return np.sum(dist_0 * dist_0, axis=1)

    return sdist


def calc_rdf(coords, cutoff=20.0, px=True, py=True, pz=True):
    """Calculate the 3D periodic radial distribution function of array of points.

    Args:
        coords (np.ndarray): Nx3 numpy array representing points in 3D
        cutoff (float): cutoff distance for neighbor consideration
        px (bool): whether or not coordinates are periodic in x
        py (bool): whether or not coordinates are periodic in y
        pz (bool): whether or not coordinates are periodic in z

    Returns:
        np.ndarray: radial distribution function, returned in steps of 0.1

    """
    # shift outside of negative octants
    coords -= np.min(coords, axis=0)

    # get box size and number of voxels in each direction
    dims = np.max(coords, axis=0)
    N = np.ceil(dims / cutoff).astype(int)

    Nt = np.array([1, N[0], N[0] * N[1]])

    # get voxel neighbor list and 1D voxel idx for each obj
    nn = get_voxel_grid(N, px, py, pz)
    box_idx = (np.floor(coords / cutoff) @ Nt).astype(int)

    # get periodic distance calculation
    f_sdist = get_sdist_fun(dims, px, py, pz)

    # assign coords to voxels
    parts = []
    part_ids = np.arange(coords.shape[0])
    for i in range(np.prod(N)):
        parts.append([int(x) for x in part_ids[box_idx == i]])

    # do 3D arrays so that distances are broadcasted/batched
    counts = np.zeros(int(cutoff / 0.1))

    # for each voxel, calculate distance of owned particles to all possible
    # neighbors within cutoff distance
    for i in range(np.prod(N)):
        cur_parts = coords[parts[i], :][:, :, None]

        for n in nn[box_idx[i]]:
            neighbor_parts = (coords[parts[n], :].T)[None, :, :]
            dist = np.sqrt(f_sdist(cur_parts, neighbor_parts))

            # use histogram to get counts for RDF
            tmp_counts, _ = np.histogram(dist,
                                         bins=counts.shape[0],
                                         range=(0.0, cutoff))
            counts += tmp_counts

    # correct for self interactions
    counts[0] = 0.0

    return counts



from abc import ABC, abstractmethod
from typing import Generator, List, Tuple

import numpy as np
from scipy.optimize import linprog
from scipy.spatial import HalfspaceIntersection

from ..util.misc import round_away

#####################################
##### Geometric Surface Classes #####
#####################################


class BaseAlgebraic(ABC):
    """Base class for algebraic surfaces.


    Attributes:
        params (Tuple): parameters describing algebraic object
        tol (float): numerical tolerance used to pad interiority checks.
                    Default is 1e-5.
    
    """

    def __init__(self, tol: float = 1e-5):
        self.tol = tol

    @abstractmethod
    def checkIfInterior(self, testPoints: np.ndarray):
        """Check if points lie on interior side of geometric surface.

        Args:
            testPoints (np.ndarray): Nx3 array of points to check.
        
        Returns:
            Nx1 logical array indicating whether or not point is on interior 
            of surface.
        """
        pass

    @property
    @abstractmethod
    def params(self):
        pass

    @property
    def tol(self):
        return self._tol

    @tol.setter
    def tol(self, val):
        assert (float(val) > 0.0)
        self._tol = float(val)


class Sphere(BaseAlgebraic):
    """Algebraic surface for spheres. 
    
    Interior points are points with distance from center smaller than the radius.
    
    Attributes:
        radius (float): Radius of sphere.
        center (np.ndarray): 3x1 array representing center of sphere in space.
        tol (float): Tolerance value for interiority check. Default is 1e-5.
    """

    def __init__(self,
                 radius: float = None,
                 center: np.ndarray = None,
                 tol=1e-5):
        self._radius = None
        self._center = np.array([0, 0, 0])

        if not (radius is None):
            self.radius = radius

        if not (center is None):
            self.center = center

        super().__init__(tol=tol)

    def checkIfInterior(self, testPoints: np.ndarray) -> np.ndarray:
        return np.sum((testPoints - self.center)**2.0,
                      axis=1) < (self.radius + self.tol)**2.0

    @property
    def params(self):
        """Return radius, center of Sphere."""
        return self.radius, self.center

    @property
    def radius(self):
        """Radius of sphere."""
        return self._radius

    @radius.setter
    def radius(self, radius: float):
        if radius > 0.0:
            self._radius = radius
        else:
            raise ValueError("Radius needs to be positive valued.")

    @property
    def center(self):
        """Center of sphere in space."""
        return self._center

    @center.setter
    def center(self, center: np.ndarray):
        center = np.array(center)  #cast to np array if not already
        assert (center.size == 3), "Center must be a point in 3D space"
        assert (center.shape[0] == 3), "Center must be a point in 3D space"
        self._center = center


class Plane(BaseAlgebraic):
    """Algebraic surface for planes in R3.

    Interior points lie opposite in direction of plane normal.

    Attributes:
        point (np.ndarray): point lying on plane.
        normal (np.ndarray): normal vector describing orientation of plane.
        tol (float): Tolerance value for interiority check. Default is 1e-5.
    
    """

    def __init__(self,
                 normal: np.ndarray = None,
                 point: np.ndarray = None,
                 tol: float = 1e-5):
        self._normal = None
        self._point = None

        if not (normal is None):
            self.normal = normal

        if not (point is None):
            self.point = point

        super().__init__(tol=tol)

    @property
    def params(self):
        """Return normal vector, point on plane of Plane."""
        return self.normal, self.point

    @property
    def point(self):
        """Point lying on surface of Plane."""
        return self._point

    @point.setter
    def point(self, point: np.ndarray):
        point = np.squeeze(np.array(point))  #cast to np array if not already
        assert (point.shape[0] == 3), "Point must be a point in 3D space"
        self._point = point

    @property
    def normal(self):
        """Normal vector defining orientation of Plane in space."""
        return self._normal

    @normal.setter
    def normal(self, normal: np.ndarray):
        normal = np.array(normal)  #cast to np array if not already
        assert (normal.size == 3), "normal must be a vector in 3D space"
        # normal = np.reshape(normal, (3,1)) #make a consistent shape
        if (np.linalg.norm(normal) > np.finfo(float).eps):
            self._normal = normal / np.linalg.norm(normal)
        else:
            raise ValueError("Normal vector must have some length")

    def checkIfInterior(self, testPoints: np.ndarray):
        return np.sum((testPoints * self.normal), axis=1) - np.squeeze(
            np.dot(self.normal, self.point + self.normal * self.tol)) < 0

    def flip_orientation(self):
        """Flip the orientation of the plane."""
        self.normal = -1 * self.normal

    def dist_from_plane(self, point: np.ndarray):
        """Calculate the distance from a point or series of points to the Plane.
        
        Arg:
            point (np.ndarray): Point in space to calculate distance.

        Returns:
            Array of distances to plane.

        """
        return np.abs(np.dot(point - self.point, self.normal))

    def project_point(self, point: np.ndarray):
        """Project a point in space onto Plane.
        
        Arg:
            point (np.ndarray): Point in space to project onto Plane.

        Returns:
            Projected point lying on surface of Plane.
        """
        return point - self.dist_from_plane(point) * self.normal.T


class Cylinder(BaseAlgebraic):
    """Algebraic surface for circular cylinders in R3.

    Cylinders are defined with vectors, pointing parallel to central axis;
    points, lying along central axis; and radii, defining size of cylinder.

    Attributes:
        axis (np.ndarray): vector parallel to central axis of cylinder.
        point (np.ndarray): point which lies along central axis of cylinder.
        radius (float): radius of cylinder.
        tol (float): Tolerance value for interiority check. Default is 1e-5.
    """

    #TODO: Write transformations for handling cylinder parameters.

    def __init__(self,
                 axis: np.ndarray = [0, 0, 1],
                 point: np.ndarray = [0, 0, 0],
                 radius: float = 1.0,
                 tol: float = 1e-5):
        self.axis = axis
        self.point = point
        self.radius = radius
        super().__init__(tol=tol)

    def params(self):
        """Return axis, point, and radius of cylinder."""
        return self.axis, self.point, self.radius

    @property
    def axis(self):
        """Vector lying parallel to central axis."""
        return self._axis

    @axis.setter
    def axis(self, arr):
        arr = np.squeeze(np.array(arr))
        assert (arr.shape == (3,))
        self._axis = arr / np.linalg.norm(arr)

    @property
    def point(self):
        """Point lying along central axis."""
        return self._point

    @point.setter
    def point(self, arr):
        arr = np.squeeze(np.array(arr))
        assert (arr.shape == (3,))
        self._point = arr

    @property
    def radius(self):
        """Radius of cylinder."""
        return self._radius

    @radius.setter
    def radius(self, val):
        try:
            val = float(val)
            self._radius = val
        except TypeError:
            raise TypeError("Supplied value must be castable to float")

    def checkIfInterior(self, testPoints):
        dists = np.linalg.norm(np.cross(testPoints - self.point,
                                        self.axis[None, :]),
                               axis=1)
        return dists < self.radius + self.tol


#####################################
######### Utility routines ##########
#####################################


def get_bounding_box(planes: List[Plane]):
    """Get convex region interior to set of Planes, if one exists.

    Determines if set of planes forms a valid interior convex region. If so, 
    returns vertices of convex region. Uses scipy half space intersection and 
    linear progamming routines to determine boundaries of convex region and 
    valid interior points.

    Args:
        planes (List[Plane]): set of planes to check mutual intersection of.

    Returns:
        np.ndarray: Nx3 array of vertices of convex region.
        2: no valid intersection.
        3: if intersection is unbounded.
    """

    # some issues arose when all planes were in negative coordinate space
    # so shift planes so that all points line positive coordinates
    shift = np.zeros_like(planes[0].point)
    for plane in planes:
        shift = np.min([shift, plane.point], axis=0)

    shift = -1.5 * shift

    # convert plane set to matrix form of half spaces
    A = np.zeros((len(planes), 3))
    d = np.zeros((len(planes), 1))
    norms = np.zeros((len(planes), 1))
    for i, plane in enumerate(planes):
        n, p = plane.params
        A[i, :] = n.squeeze()
        d[i, 0] = -1.0 * np.dot(n.squeeze(), (p + shift).squeeze())
        norms[i, 0] = np.linalg.norm(n)

    #check feasiblity of region and get interior point
    c = np.zeros(4)
    c[-1] = -1
    res = linprog(c, A_ub=np.hstack([A, norms]), b_ub=-1.0 * d)

    if (res.status == 0):
        hs = HalfspaceIntersection(np.hstack([A, d]), res.x[:-1])
        return hs.intersections - shift
    else:
        return res.status


def snap_plane_near_point(point: np.ndarray,
                          generator: Generator,
                          miller_indices: Tuple[int],
                          mode: str = "nearest"):
    """Determine nearest crystallographic nearest to point in space for given crystal coordinate system.
    
    Args:
        point (np.ndarray): Point in space.
        generator (Generator): Generator describing crystal coordinat system.
        miller_indices (Tuple[int]): miller indices of desired plane.
        mode (str): "nearest" for absolute closest plane to point; "floor" for
                    next nearest valid plane towards generator origin; "ceil"
                    for next furthest valid plane from generator origin.

    Returns:
        Plane in space with orientation given by Miller indices snapped to 
        nearest valid location.

    """

    miller_indices = np.array(miller_indices)

    # get point coordinates in generator coordinate system
    point_fcoord = np.array(np.linalg.solve(generator.voxel.sbases, point))

    # get lattice points that are intersected by miller plane
    with np.errstate(divide="ignore"):  #check for infs directly
        target_fcoord = 1 / miller_indices

    new_point = np.zeros((3, 1))

    # TODO: if bases are not orthonormal, this procedure is not correct
    # since the following rounds towards the nearest lattice points, with equal
    # weights given to all lattice vectors
    if mode == "nearest":
        for i in range(3):
            new_point[i,0] = np.round(point_fcoord[i]/target_fcoord[i])*target_fcoord[i] \
                                if not np.isinf(target_fcoord[i]) else point_fcoord[i]
    elif mode == "ceil":
        for i in range(3):
            new_point[i,0] = round_away(point_fcoord[i]/target_fcoord[i])*target_fcoord[i] \
                                if not np.isinf(target_fcoord[i]) else point_fcoord[i]
    elif mode == "floor":
        for i in range(3):
            new_point[i,0] = np.fix(point_fcoord[i]/target_fcoord[i])*target_fcoord[i] \
                                if not np.isinf(target_fcoord[i]) else point_fcoord[i]

    # scale back to real space
    new_point = generator.voxel.sbases @ new_point

    # get perpendicular vector
    normal = generator.voxel.reciprocal_bases.T @ miller_indices
    return Plane(normal=normal, point=new_point)



import copy
from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from ..generator import Generator
from ..transform import *
from ..util.misc import get_N_splits
from ..volume import BaseVolume, MultiVolume, Volume
from ..volume.algebraic import Plane, snap_plane_near_point


class BasePrefab(ABC):
    """Base abstract class for Prefab objects.

    Prefab objects are objects and classes that can run predesigned algorithms
    for generating certain classes of regular objects, typically with 
    sampleable features or properties. For example, planar defects in FCC 
    systems are easily described in algorithmic form-- a series of {111} planes 
    can be chosen to put a defect on.

    Prefab objects will generally take in at least a base Generator object defining
    the system of interest and potentially take in Volume objects. They will return
    Volumes, or, more likely, MultiVolume objects which contains the resultant
    structure defined by the prefab routine.
    """

    @abstractmethod
    def build_object(self) -> BaseVolume:
        """Construct and return a prefabicated structure."""
        pass


class fccMixedTwinSF(BasePrefab):
    """Prefab routine for returning FCC volume with mixed stacking fault and twin defects.

    Given a volume with an attached generator (assumed to be FCC symmetry), 
    return a MultiVolume with a series of twin defects and stacking faults. 
    Defects can be placed on any plane in the {111} family, and are uniformly
    randomly distributed on the available (111) planes with a minimum separation.
    Types of defects are uniformly randomly sampled, as well, according to a ratio
    of twin defects to stacking faults.

    Attributes:
        N (int): Number of defects to attempt to place into volume
        plane (Tuple[int]): Length 3 tuple of Miller indices representing set of 
                            planes to put defects on
        ratio (float): Ratio of stacking faults:total defects. A ratio of 1.0 
                        will produce only stacking faults, while a ratio of 0.0
                        will produce only twin defects. 
        generator (Generator): Generator object representing FCC crystal system.
        volume (Volume): Volume object representing bounds of object in which
                        defects will be placed.

    """

    def __init__(self,
                 generator: Generator = None,
                 volume: BaseVolume = None,
                 ratio: float = 0.5,
                 N: int = 1,
                 min_sep: int = 3,
                 plane: Tuple[int] = (1, 1, 1)):
        self._N = None
        self._plane = None
        self._ratio = None
        self._min_sep = None
        self._generator = None
        self._volume = None

        self.N = N
        self.plane = plane
        self.ratio = ratio
        self.min_sep = min_sep

        if not generator is None:
            self.generator = generator

        if not volume is None:
            self.volume = volume

    @property
    def N(self):
        """Number of defects to place in volume."""
        return self._N

    @N.setter
    def N(self, val):
        self._N = int(val)

    @property
    def min_sep(self):
        """Minimum seperation between defects in numbers of planes."""
        return self._min_sep

    @min_sep.setter
    def min_sep(self, val: int):
        assert (
            isinstance(val, int)
        ), "Must supply integer number of planes for minimum seperation of defects."
        self._min_sep = val

    @property
    def plane(self):
        """Miller indices of defect planes."""
        return self._plane

    @plane.setter
    def plane(self, val: Tuple[int]):
        self._plane = val

    @property
    def ratio(self):
        """Ratio of stacking faults:total defects placed into object."""
        return self._ratio

    @ratio.setter
    def ratio(self, val: float):
        assert (0.0 <= val and
                val <= 1.0), "Ratio must be value in interval [0,1)."
        self._ratio = val

    @property
    def generator(self):
        """Crystalline geneartor used to sample planes for defects."""
        return self._generator

    @generator.setter
    def generator(self, val: Generator):
        assert (isinstance(
            val, Generator)), "Must supply crystalline Generator object."
        self._generator = val

    @property
    def volume(self):
        """Volume used to defined outer bounds of defected object."""
        return self._volume

    @volume.setter
    def volume(self, val: BaseVolume):
        assert (isinstance(
            val,
            BaseVolume)), "Must supply either Volume or MultiVolume object."
        self._volume = val

    def build_object(self):
        # get list of all planes in bounding box
        # TODO: the bounding box isn't necessarily tangent to the valid volume (e.g., spheres)
        # perhaps refine the end points until planes intersect
        norm_vec = self.generator.voxel.bases @ np.array(
            self.plane)  # fine to use real bases since cubic
        norm_vec /= np.linalg.norm(norm_vec)
        bbox = self.volume.get_bounding_box()
        ci = np.dot(norm_vec.reshape(1, 3), bbox.T).T
        start = bbox[np.argmin(ci), :]
        finish = bbox[np.argmax(ci), :]

        if isinstance(self.volume.alg_objects[0], Sphere):
            # override bbox in this special case
            start = -1.0 * self.volume.alg_objects[0].radius * norm_vec
            finish = self.volume.alg_objects[0].radius * norm_vec

        sp = snap_plane_near_point(start,
                                   self.generator,
                                   self.plane,
                                   mode="floor")
        ep = snap_plane_near_point(finish,
                                   self.generator,
                                   self.plane,
                                   mode="ceil")
        d_tot = ep.dist_from_plane(sp.point)
        d_hkl = self.generator.lattice.d_hkl(self.plane)
        N_planes = np.round(d_tot / d_hkl).astype(int)
        planes = [sp]
        new_point = np.copy(sp.point)
        for i in range(N_planes):
            new_point += sp.normal * d_hkl
            planes.append(Plane(normal=sp.normal, point=new_point))

        # select N planes with min separation apart
        splits = get_N_splits(self.N, self.min_sep, len(planes))
        splits.reverse()

        # create sub volumes for final multivolume
        # origins should be successively shifted
        gen_tmp = self.generator.from_generator()
        vols = [self.volume.from_volume(generator=gen_tmp)]
        vols[0].priority = self.N
        twin_last = 1
        for i in range(self.N):
            if (np.random.rand() < self.ratio):
                # add stacking fault
                burger = twin_last * gen_tmp.voxel.sbases @ (
                    (1 / 3) * np.array([1, 1, -2]) * np.sign(self.plane))
                t = Translation(shift=burger)
            else:
                # add twin defect
                t = Reflection(planes[splits[i]])
                twin_last *= -1
            gen_tmp = gen_tmp.from_generator(transformation=[t])
            new_vol = self.volume.from_volume(generator=gen_tmp)
            plane_tmp = Plane(normal=1.0 * planes[splits[i]].normal,
                              point=planes[splits[i]].point)
            new_vol.add_alg_object(plane_tmp)
            new_vol.priority = self.N - (i + 1)
            vols.append(new_vol)

        return MultiVolume(volumes=vols)


class fccStackingFault(fccMixedTwinSF):
    """Prefab routine for returning FCC volume with stacking fault defects.

    Given a volume with an attached generator (assumed to be FCC symmetry), 
    return a MultiVolume with a series of. stacking faults. Defects can be placed
    on any plane in the {111} family, and are uniformly randomly distributed on 
    the available (111) planes with a minimum separation.

    Attributes:
        N (int): Number of defects to attempt to place into volume
        plane (Tuple[int]): Length 3 tuple of Miller indices representing set of 
                            planes to put defects on/
        generator (Generator): Generator object representing FCC crystal system.
        volume (Volume): Volume object representing bounds of object in which
                        defects will be placed.
    """

    def __init__(self,
                 generator: Generator = None,
                 volume: BaseVolume = None,
                 N: int = 1,
                 min_sep: int = 3,
                 plane: Tuple[int] = (1, 1, 1)):
        super().__init__(generator=generator,
                         volume=volume,
                         N=N,
                         min_sep=min_sep,
                         plane=plane)

    @property
    def ratio(self):
        return 1.0


class fccTwin(fccMixedTwinSF):
    """Prefab routine for returning FCC volume with twin defects.

    Given a volume with an attached generator (assumed to be FCC symmetry), 
    return a MultiVolume with a series of. stacking faults. Defects can be placed
    on any plane in the {111} family, and are uniformly randomly distributed on 
    the available (111) planes with a minimum separation.

    Attributes:
        N (int): Number of defects to attempt to place into volume
        plane (Tuple[int]): Length 3 tuple of Miller indices representing set of 
                            planes to put defects on/
        generator (Generator): Generator object representing FCC crystal system.
        volume (Volume): Volume object representing bounds of object in which
                        defects will be placed.
    """

    def __init__(self,
                 generator: Generator = None,
                 volume: BaseVolume = None,
                 N: int = 1,
                 min_sep: int = 3,
                 plane: Tuple[int] = (1, 1, 1)):
        super().__init__(generator=generator,
                         volume=volume,
                         N=N,
                         min_sep=min_sep,
                         plane=plane)

    @property
    def ratio(self):
        return 0.0


class SimpleGrainBoundary(BasePrefab):
    """Prefab routine for crystalline grain boundaries.

    Under development.
    """

    def __init__(self,
                 z1,
                 r1,
                 z2=None,
                 r2=None,
                 plane=None,
                 point=None,
                 volume=None,
                 generator=None):
        """
        z1, z2 are zone axes for grains 1, 2
        r1 and r2 are phase about the zone axes, respectively
        plane is either a Plane object defining a split plane or miller indices of split plane, nearest center of volume
        """
        self.z1 = z1
        self.z2 = z2
        self.r1 = r1
        self.r2 = r2
        self.plane = plane
        self.volume = volume
        self.generator = generator

    @property
    def z1(self):
        return self._z1

    @z1.setter
    def z1(self, z1):
        z1 = np.array(z1)
        assert (
            z1.size == 3
        ), "Zone axes inputs must be 3-element list, tuple, or numpy array"
        self._z1 = z1

    @property
    def z2(self):
        return self._z2

    @z2.setter
    def z2(self, z2):
        if z2 is None:
            self._z2 = self.z1
        else:
            z2 = np.array(z2)
            assert (
                z2.size == 3
            ), "Zone axes inputs must be 3-element list, tuple, or numpy array"
            self._z2 = z2

    def build_object(self):

        zv = np.array([0, 0, 1])
        z1v = self.generator.voxel.sbases @ self.z1
        z2v = self.generator.voxel.sbases @ self.z2

        rot_1 = Rotation(rot_vtv(z1v, zv))
        rot_2 = Rotation(rot_vtv(z2v, zv))
        gen_1 = self.generator.from_generator(transformation=[rot_1])
        gen_2 = self.generator.from_generator(transformation=[rot_2])

        vol_1 = self.volume.from_volume(generator=gen_1)
        vol_1.add_alg_object(self.plane)
        vol_2 = self.volume.from_volume(generator=gen_2)
        vol_2.priority += 1

        return MultiVolume(volumes=[vol_1, vol_2])



import copy
import itertools
from abc import ABC, abstractmethod
from typing import List, Tuple

import numpy as np
from pymatgen.core import Lattice, Structure
from pymatgen.symmetry.groups import SpaceGroup, sg_symbol_from_int_number

from ..transform import BaseTransform
from ..transform.strain import BaseStrain
from ..volume.voxel import Voxel
from .amorphous_algorithms import *

#####################################
########## Generator Classes ########
#####################################


class BaseGenerator(ABC):
    """Base abstract class for Generator objects. 
    
    Generator objects are additive components in Construction Zone. When designing
    nanostructures, Generators contain information about the arrangement of atoms
    in space and can supply atoms at least where they should exist. 

    BaseGenerators are typically not created directly. Use the Generator class
    for crystalline systems, and the AmorphousGenerator class for non-crystalline
    systems.
    """

    @abstractmethod
    def supply_atoms(self, bbox: np.ndarray):
        """Given a bounding region, supply enough atoms to complete fill the region.
        
        Args:
            bbox (np.ndarray): Nx3 array defining vertices of convex region

        Returns:
            Coordinates and species of atoms that fill convex region. 
            Returned as Nx3 and Nx1 arrays.
        """
        pass

    def from_generator(self, **kwargs):
        """Constructor for new Generators based on existing Generator object.

        Args:
            **kwargs: "transformation"=List[BaseTransformation] to apply a 
                        series of transformations to the copied generator.
    
        """
        new_generator = copy.deepcopy(self)

        if "transformation" in kwargs.keys():
            for t in kwargs["transformation"]:
                new_generator.transform(t)

        return new_generator


class Generator(BaseGenerator):
    """Generator object for crystal systems.

    Generator objects are additive components in Construction Zone. When designing
    nanostructures, Generators contain information about the arrangement of atoms
    in space and can supply atoms at least where they should exist. 

    The Generator class handles crystalline systems, primarily by utilizing 
    Structure and Lattice objects from pymatgen, along with an internal Voxel 
    class that maintains the state of the Generator if transformed.

    Attributes:
        structure (Structure): pymatgen Structure object encoding crystallographic
                                details.
        lattice (Lattice): Convenience property to grab lattice information from
                            Structure.
        species (np.ndarray): Atomic numbers for atoms in unit cell.
        coords (np.ndarray): Fractional coordinates of atoms in unit cell.
        orientation (np.ndarray): Orientation of generator. Not used currently.
        origin (np.ndarray): Local origin of generator in Cartesian coordinates.
        voxel (Voxel): Internal Voxel object used to span space for generator.
        strain_field (BaseStrain): Strain field applied to atoms supplied by Generator.

    """

    def __init__(self,
                 origin: np.ndarray = None,
                 structure: Structure = None,
                 strain_field: BaseStrain = None):
        self._structure = None
        self._voxel = None
        self._orientation = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
        self._strain_field = None

        if not structure is None:
            self.structure = structure

        if not strain_field is None:
            self.strain_field = strain_field

        if not origin is None:
            self.origin = origin

    """
    Properties
    """

    @property
    def structure(self):
        """pymatgen Structure object encoding crystallographic details."""
        return self._structure

    @structure.setter
    def structure(self, structure: Structure):
        assert (isinstance(
            structure,
            Structure)), "Structure must be pymatgen Structure object"
        self._structure = structure

        # check if voxel exists, if so, copy origin
        if self.voxel is None:
            self.voxel = Voxel(bases=structure.lattice.matrix)
        else:
            self.voxel = Voxel(bases=structure.lattice.matrix,
                               origin=self.origin)

    @property
    def lattice(self):
        """Crystallographic lattice and unit cell information."""
        return self.structure.lattice

    @property
    def species(self):
        """Ordered list of species in unit cell."""
        return self.structure.atomic_numbers

    @property
    def coords(self):
        """Ordered list of fractional coordinates of atoms in unit cell."""
        return self.structure.frac_coords

    @property
    def orientation(self):
        """Orientation of unit cell in space. Not used."""
        return self._orientation

    @property
    def origin(self):
        """Local origin of Generator in Cartesian coordinates."""
        return self.voxel.origin

    @origin.setter
    def origin(self, val):
        if self.voxel is None:
            # create voxel to store origin for when structure is later attached
            self.voxel = Voxel(origin=val)
        else:
            self.voxel.origin = val

    @property
    def voxel(self):
        """Voxel object used to span space and keep transformed bases for Generator."""
        return self._voxel

    @voxel.setter
    def voxel(self, voxel: Voxel):
        if not isinstance(voxel, Voxel):
            raise TypeError("Supplied voxel is not of Voxel() class")

        self._voxel = voxel

    @property
    def strain_field(self):
        """Strain field applied to atoms post-generation."""
        return self._strain_field

    @strain_field.setter
    def strain_field(self, field: BaseStrain):
        assert (isinstance(
            field, BaseStrain)), "Strain field must be of BaseStrain class"
        self._strain_field = field

    """ 
    Methods
    """

    def supply_atoms(self, bbox: np.ndarray):
        #generation method for pymatgen lattice structure
        min_extent, max_extent = self.voxel.get_extents(bbox)
        fcoords = np.copy(
            self.coords)  #grab fractional in case bases are rotated

        # get matrix representing unit cells as grid
        l = [
            range(min_extent[0], max_extent[0] + 1),
            range(min_extent[1], max_extent[1] + 1),
            range(min_extent[2], max_extent[2] + 1), 
            [0]
        ]
        ucs = np.array(list(itertools.product(*l)))

        # get 4D bases with identity in final index to multiply species list
        bases = np.eye(4)
        bases[0:3, 0:3] = self.voxel.sbases
        lcoords = bases @ ucs.T
        scoords = bases @ (np.hstack([fcoords,
                                      np.array(self.species)[:, None]])).T

        coords = lcoords[:, :, None] + scoords[:, None, :]
        coords = coords.reshape(4, coords.shape[1] * coords.shape[2])
        coords = coords.T

        out_coords = coords[:, :-1] + self.voxel.origin
        species = coords[:, -1]

        if self.strain_field is None:
            return out_coords, species
        else:
            self.strain_field.scrape_params(self)
            return self.strain_field.apply_strain(out_coords), species

    def transform(self, transformation: BaseTransform):
        """Transform Generator object with transformation described by Transformation object.
        
        Args:
            transformation (BaseTransform): Transformation object from transforms module.
        """
        assert (isinstance(transformation, BaseTransform)
               ), "Supplied transformation not transformation object."
        self.voxel.bases = transformation.applyTransformation_bases(
            self.voxel.bases)

        if not (transformation.basis_only):
            new_origin = transformation.applyTransformation(
                np.reshape(self.voxel.origin, (1, 3)))
            self.voxel.origin = np.squeeze(new_origin)

    @classmethod
    def from_spacegroup(cls,
                        Z: List[int],
                        coords: np.ndarray,
                        cellDims: np.ndarray,
                        cellAngs: np.ndarray,
                        sgn: int = None,
                        sym: str = None,
                        **kwargs):
        """Convenience constructor for creating Generator with symmetric unit cell from spacegroup.
        
        Args:
            Z (List[int]): List of atomic numbers in symmetric sites of unit cell.
            coords (np.ndarray): Nx3 array of fractional coordinates of atoms
                                at symmetric sites in unit cell.
            cellDims (List[float]): Length 3 list with lattice parameters a,b,c of unit cell.
            cellAngs (List[float]): Length 3 list with lattice parameters alpha,
                                    beta, gamma of unit cell.
            sgn (int): Space group number of desired space group.
            sym (str): Space group symbol of desired space group as Full International
                        or Hermann-Mauguin symbol. Overriden by sgn, if both supplied.
            **kwargs: Any of **kwargs accepted in standard Generator construction.

        Returns:
            Generator: Generator object with structure and symmetry defined by input space group.
        """
        if sym is None:
            sg = SpaceGroup(int_symbol=sg_symbol_from_int_number(sgn))
        else:
            sg = SpaceGroup(int_symbol=sym)

        lattice = Lattice.from_parameters(a=cellDims[0],
                                          b=cellDims[1],
                                          c=cellDims[2],
                                          alpha=cellAngs[0],
                                          beta=cellAngs[1],
                                          gamma=cellAngs[2])

        structure = Structure.from_spacegroup(sg.int_number,
                                              lattice=lattice,
                                              species=Z,
                                              coords=coords)

        return cls(structure=structure, **kwargs)

    @classmethod
    def from_unit_cell(cls, Z: List[int], coords: np.ndarray,
                       cellDims: np.ndarray, cellAngs: np.ndarray, **kwargs):
        """Convenience constructor for creating Generators directly from unit cell data.

        Args:
            Z (List[int]): List of atomic numbers in symmetric sites of unit cell.
            coords (np.ndarray): Nx3 array of fractional coordinates of atoms
                                at symmetric sites in unit cell.
            cellDims (List[float]): Length 3 list with lattice parameters a,b,c of unit cell.
            cellAngs (List[float]): Length 3 list with lattice parameters alpha,
                                    beta, gamma of unit cell.
            **kwargs: Any of **kwargs accepted in standard Generator construction.

        Returns:
            Generator: Generator object with given unit cell.
        """
        tmp_lattice = Lattice.from_parameters(a=cellDims[0], b=cellDims[1], c=cellDims[2], \
                                                alpha=cellAngs[0], beta=cellAngs[1], gamma=cellAngs[2])

        structure = Structure(tmp_lattice, Z, coords)

        return cls(structure=structure, **kwargs)


class AmorphousGenerator(BaseGenerator):
    """Generator object for non-crystal systems.

    Generator objects are additive components in Construction Zone. When designing
    nanostructures, Generators contain information about the arrangement of atoms
    in space and can supply atoms at least where they should exist. 

    The AmorphousGenerator object handles non-crystalline systems.
    Currently supports only monatomic, periodically uniformly disrtributed blocks.
    Default parameters are set for the purpose of generating amorphous carbon 
    blocks.

    Attributes:
        origin (np.ndarray): Local origin of generator in Cartesian coordinates.
        species (int): Atomic number of element used in monatomic generation.
        density (float): Average density of material.
        min_dist (float): Minimum bond distance between atoms.
        old_result (np.ndarray): Coordinates of atoms in previous generation.
        use_old_result (bool): Whether or not to re-use previous generation, or 
                               to run generation routine again.

    """

    def __init__(self, origin=None, min_dist=1.4, density=.1103075, species=6):
        self._origin = None
        self._species = None
        self._density = None
        self._min_dist = None

        if not (origin is None):
            self.origin = origin

        self.species = species
        self.min_dist = min_dist
        self.density = density
        self.use_old_result = False

    """
    Properties
    """

    @property
    def origin(self):
        """Local origin of Generator in Cartesian coordinates."""
        return self._origin

    @origin.setter
    def origin(self, origin: np.ndarray):
        origin = np.array(origin)
        assert (origin.size == 3), "Origin must be a point in 3D space"
        self._origin = origin

    @property
    def species(self):
        """Atomic species used in monatomic generation."""
        return self._species

    @species.setter
    def species(self, species):
        self._species = species

    @property
    def density(self):
        """Density of atoms in material."""
        return self._density

    @density.setter
    def density(self, density):
        assert (density > 0.0)
        self._density = density

    @property
    def min_dist(self):
        """Minimum bond distance between atoms in material."""
        return self._min_dist

    @min_dist.setter
    def min_dist(self, min_dist):
        assert (min_dist > 0)
        self._min_dist = min_dist

    @property
    def old_result(self):
        """Atomic coordinates of previous generation, if supply atoms has been called."""
        return self._old_result

    """
    Methods
    """

    def supply_atoms(self, bbox):

        if self.use_old_result and self._old_result is not None:
            return self.old_result
        else:
            coords = gen_p_substrate(
                np.max(bbox, axis=0) - np.min(bbox, axis=0), self.min_dist)
            self._old_result = (coords, np.ones(coords.shape[0]) * self.species)
            return self.old_result



from setuptools import setup, find_packages

setup(
    name='czone',
    version='2021.08.04',
    description='An open source python package for generating nanoscale+ atomic scenes',
    url='https://github.com/lerandc/construction_zone',
    author='Luis Rangel DaCosta',
    author_email='luisrd@berkeley.edu',
    python_requires='>=3.7',
    packages=find_packages(),
    install_requires=[
        'pymatgen >= 2020.4.29',
        'numpy >= 1.16.2',
        'scipy >= 1.3.0',
        'ase >= 3.21.0',
        'wulffpack >= 1.1.0']
)


import copy
from abc import ABC, abstractmethod

import numpy as np
from scipy.spatial.transform import \
    Rotation as scRotation  # avoid namespace conflicts

from ..volume.algebraic import BaseAlgebraic, Plane, Sphere

#####################################
########### Base Classes ############
#####################################


class BaseTransform(ABC):
    """Base class for transformation objects which manipulate Generators and Volumes.

    Transformation objects contain logic and parameters for manipulating the 
    different types of objects used in Construction Zone, namely, Generators and
    Volumes. BaseTransform is typically not created directly. Use MatrixTransform for 
    generalized matrix transformations.

    Attributes:
        locked (bool): whether or not transformation applies jointly to volumes containing generators
        basis_only (bool): whether or not transformation applies only to basis of generators
        params (tuple): parameters describing transformation

    """

    def __init__(self, locked: bool = True, basis_only: bool = False):
        self.locked = locked
        self.basis_only = basis_only

    @abstractmethod
    def applyTransformation(self, points: np.ndarray) -> np.ndarray:
        """Apply transformation to a collection of points in space.
     
        Args:
            points (np.ndarray): Nx3 array of points to transform.

        Returns:
            np.ndarray: Nx3 array of transformed points.
        """
        pass

    @abstractmethod
    def applyTransformation_bases(self, points: np.ndarray) -> np.ndarray:
        """Apply transformation to bases of a generator.

        Args:
            points (np.ndarray): 3x3 array of bases from generator voxel.

        Returns:
            np.ndarray: transformed 3x3 array of bases.
        """
        pass

    @abstractmethod
    def applyTransformation_alg(self,
                                alg_object: BaseAlgebraic) -> BaseAlgebraic:
        """Apply transformation to algebraic object.

        Args:
            alg_object (BaseAlgebraic): Algebraic object to transform.

        Returns:
            BaseAlgebraic: Transformed object.
        """
        pass

    @property
    @abstractmethod
    def params(self) -> tuple:
        """Return parameters describing transformation. """
        pass

    @property
    def locked(self) -> bool:
        """Boolean value indicating whether or not transformation jointly applied to Volumes and Generators."""
        return self._locked

    @locked.setter
    def locked(self, locked):
        assert (isinstance(locked,
                           bool)), "Must supply bool to locked parameter."
        self._locked = locked

    @property
    def basis_only(self) -> bool:
        """Boolean value indicating whether or not transformation applied only to basis of Generators."""
        return self._basis_only

    @basis_only.setter
    def basis_only(self, basis_only):
        assert (isinstance(basis_only,
                           bool)), "Must supply bool to basis_only parameter"
        self._basis_only = basis_only


class Translation(BaseTransform):
    """Transformation object that applies translations to Generators and Volumes.

    Attributes:
        shift (np.ndarray): Translation vector in 3D space.
    """

    def __init__(self,
                 shift=None,
                 locked: bool = True,
                 basis_only: bool = False):
        self._shift = None
        self._locked = None

        if not (shift is None):
            self.shift = shift

        self.locked = locked
        self.basis_only = basis_only

    @property
    def shift(self):
        return self._shift

    @shift.setter
    def shift(self, shift):
        shift = np.array(shift)
        assert (shift.size == 3), "Shift must be a vector in 3D space"
        assert (shift.shape[0] == 3), "Shift must be a vector in 3D space"
        self._shift = shift

    @property
    def params(self):
        return self.shift

    def applyTransformation(self, points):
        return points + self.shift

    def applyTransformation_bases(self, points):
        return points

    def applyTransformation_alg(self, alg_object):
        if isinstance(alg_object, Sphere):
            alg_object.center = alg_object.center + self.shift

        if isinstance(alg_object, Plane):
            alg_object.point = alg_object.point + self.shift

        return alg_object


class MatrixTransform(BaseTransform):
    """Transformation object that applies arbitrary matrix transformations.

    Inversion, Rotation, and Reflection classes are derived classes of MatrixTransform
    that have special setters which ensure validity of their respective transformations.

    Attributes:
        matrix (np.ndarray): 3x3 matrix representing transformation
        origin (np.ndarray): point in R3 representing relative origin in which 
                             transformation is applied
    """

    def __init__(self,
                 matrix=None,
                 origin=None,
                 locked: bool = True,
                 basis_only: bool = False):
        self._matrix = None
        self._origin = None

        if not (matrix is None):
            self.matrix = matrix

        if not (origin is None):
            self.origin = origin

        super().__init__(locked=locked, basis_only=basis_only)

    @property
    def matrix(self) -> np.ndarray:
        """3x3 array representing matrix transformation """
        return self._matrix

    @matrix.setter
    def matrix(self, matrix: np.ndarray):
        assert (matrix.shape == (
            3, 3)), "Input matrix must be square 3x3 numpy array"
        self._matrix = matrix

    @property
    def origin(self) -> np.ndarray:
        """Point in R3 representing relative origin of transformation"""
        return self._origin

    @origin.setter
    def origin(self, origin):
        origin = np.array(origin)
        assert (origin.size == 3), "Origin must be a point in 3D space"
        assert (origin.shape[0] == 3), "Origin must be a point in 3D space"
        self._origin = origin

    @property
    def params(self):
        return self.matrix, self.origin

    def applyTransformation(self, points):
        if not (self.origin is None):
            points = np.dot(self.matrix,
                            (points - self.origin).T).T + self.origin
        else:
            points = np.dot(self.matrix, (points).T).T

        return points

    def applyTransformation_bases(self, points):
        return np.dot(self.matrix, points)

    def applyTransformation_alg(self, alg_object):

        if isinstance(alg_object, Sphere):
            if not (self.origin is None):
                alg_object.center = np.dot(
                    self.matrix,
                    (alg_object.center - self.origin).T).T + self.origin
            else:
                alg_object.center = np.dot(self.matrix, (alg_object.center).T).T

        if isinstance(alg_object, Plane):
            if not (self.origin is None):
                alg_object.point = np.dot(
                    self.matrix,
                    (alg_object.point - self.origin).T).T + self.origin
            else:
                alg_object.point = np.dot(self.matrix, (alg_object.point).T).T

            alg_object.normal = np.dot(self.matrix, (alg_object.normal).T).T

        return alg_object


class Inversion(MatrixTransform):
    """ Transformation which applies inversion about an origin.

    Inversion, Rotation, and Reflection classes are derived classes of MatrixTransform
    that have special setters which ensure validity of their respective transformations.
    """

    @property
    def matrix(self):
        """Inversion matrix. Cannot be altered."""
        return -1.0 * np.eye(3)


class Rotation(MatrixTransform):
    """Transformation which applies a rotation about an origin.

    Inversion, Rotation, and Reflection classes are derived classes of MatrixTransform
    that have special setters which ensure validity of their respective transformations.
    """

    def __init__(self,
                 matrix=None,
                 origin=None,
                 locked: bool = True,
                 basis_only: bool = False):
        super().__init__(matrix=matrix,
                         origin=origin,
                         locked=locked,
                         basis_only=basis_only)

    @MatrixTransform.matrix.setter
    def matrix(self, val):
        assert (val.shape == (3,
                              3)), "Input matrix must be square 3x3 numpy array"
        assert (np.abs(np.linalg.det(val) - 1.0) < 1e-6
               ), "Input matrix not a valid rotation matrix. Fails determinant."
        assert (
            np.sum(np.abs(val @ val.T - np.eye(3))) < 1e-6
        ), "Input matrix not a valid rotation matrix. Fails orthogonality."
        self._matrix = val


class Reflection(MatrixTransform):
    """Transformation which applies a relfection about an arbitrary plane.

    Inversion, Rotation, and Reflection classes are derived classes of MatrixTransform
    that have special setters which ensure validity of their respective transformations.

    Attributes:
        plane (Plane): plane about which reflection is performed
    """

    def __init__(self,
                 plane=None,
                 locked: bool = True,
                 basis_only: bool = False):
        super().__init__(matrix=None,
                         origin=None,
                         locked=locked,
                         basis_only=basis_only)

        if not (plane is None):
            self.plane = plane

    @MatrixTransform.matrix.setter
    def matrix(self, normal):
        self._matrix = np.eye(3) - (2.0 /
                                    (normal.T @ normal)) * (normal @ normal.T)

    @property
    def plane(self):
        return self._plane

    @plane.setter
    def plane(self, plane):
        assert (isinstance(
            plane,
            Plane)), "Input plane must be Plane object from algebraic module"
        self._plane = plane
        self.matrix = plane.normal.reshape(3, 1)
        self.origin = plane.point

    @property
    def params(self):
        return self.plane

    def applyTransformation_alg(self, alg_object):
        if (isinstance(alg_object), Sphere):
            alg_object.center = np.dot(
                self.matrix,
                (alg_object.center - self.origin).T).T + self.origin

        if (isinstance(alg_object), Plane):
            alg_object.point = np.dot(
                self.matrix, (alg_object.point - self.origin).T).T + self.origin

        return alg_object


class MultiTransform(BaseTransform):
    """Transformation sequence which applies a numeruous transformations in succession.

    Attributes:
        transforms (list[BaseTransform]): list of individual transformations applied
        locked (bool): whether any of transformations applied is a joint transformation
        basic_only (bool): whether any of transformations applied transforms 
                           only the generator basis

    """

    def __init__(self, transforms=None):
        self._transforms = []

        if not transforms is None:
            self.add_transform(transforms)

    @property
    def transforms(self):
        """ Sequence of individual transformations to apply"""
        return self._transforms

    @property
    def params(self):
        return [x.params for x in self.transforms]

    @property
    def locked(self):
        """Boolean value indicating whether or not any of transformations applies jointly."""
        locked_list = [x.locked for x in self.transforms]
        return np.any(locked_list)

    @property
    def basis_only(self):
        """Boolean value indicating whether or not any of transformations applies only to basis of generator"""
        """
        Essetially signals to generator that at least one of its transforms will transform the origin
        Should instead pass the voxel to the transformation objects and handle it internally per transform
        because if not all of them affect the origin, the origin will get wrongly manipulated.
        """
        basis_only_list = [x.basis_only for x in self.transforms]
        return np.any(basis_only_list)

    """
    Methods
    """

    def add_transform(self, transform):
        """Add a transformation to sequence of transformations"""
        if hasattr(transform, '__iter__'):
            for t in transform:
                assert (isinstance(t, BaseTransform)
                       ), "transforms must be BaseTransformation objects"
            self._transforms.extend(transform)
        else:
            assert (isinstance(
                transform,
                BaseTransform)), "transforms must be BaseTransformation objects"
            self._transforms.append(transform)

    def applyTransformation(self, points):
        for t in self.transforms:
            points = t.applyTransformation(points)
        return points

    def applyTransformation_bases(self, points):
        for t in self.transforms:
            points = t.applyTransformation_bases(points)
        return points

    def applyTransformation_alg(self, alg_object):
        for t in self.transforms:
            alg_object = t.applyTransformation_alg(alg_object)
        return alg_object


#####################################
############# Utilities #############
#####################################


def rot_v(v, theta):
    """Calculate rotation about an arbitary vector.

    Args:
        v (np.ndarray): axis of rotation
        theta (float): angle of rotation in radians

    Returns:
        np.ndarray: 3x3 rotation matrix
    """
    v = np.array(v)
    v = v * (theta / np.linalg.norm(v))

    return scRotation.from_rotvec(v).as_matrix()


def rot_vtv(v: np.ndarray, vt: np.ndarray) -> np.ndarray:
    """Calculate rotation to align one vector to another.

    Args:
        v (np.ndarray): 1x3 original vector
        vt (np.ndarray): 1x3 target vector into which v is rotated
    
    Returns:
        np.ndarray: 3x3 rotation matrix
    """
    """
    implementation based on following:
    https://math.stackexchange.com/questions/293116/rotating-one-3d-vector-to-another
    https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula

    In short, use cross product to get axis of rotation, then develop matrix form of
    Rodrigues rotation of formula and return rotation matrix.
    """
    v = np.array(v).reshape(3,)
    vt = np.array(vt).reshape(3,)
    eps = np.finfo(float).eps  #get machine epsilon for collinearity check
    theta = np.arccos(np.dot(v, vt) / (np.linalg.norm(v) * np.linalg.norm(vt)))
    if (theta < eps):
        return np.eye(3)
    elif (np.pi - theta < eps):
        e_i = np.eye(3)[np.argmin(v)]  #grab axis along minimum component of v
        rot_axis = np.cross(v, e_i).astype(float)
    else:
        rot_axis = np.cross(v, vt).astype(float)

    rot_axis /= np.linalg.norm(rot_axis)
    A = np.array([[0, -rot_axis[2],
                   rot_axis[1]], [rot_axis[2], 0, -rot_axis[0]],
                  [-rot_axis[1], rot_axis[0], 0]])

    return np.eye(3) + np.sin(theta) * A + (1.0 - np.cos(theta)) * A @ A


def rot_align(v: np.ndarray, vt: np.ndarray) -> np.ndarray:
    """Calculate rotation aligning two sets of ordered vectors.

    Args:
        v  (np.ndarray): Nx3 set of vectors
        vt (np.ndarray): Nx3 set of corresponding target vectors

    Returns:
        np.ndarray: 3x3 rotation matrix
    """

    return scRotation.align_vectors(vt, v).as_matrix()


def rot_zxz(alpha: float,
            beta: float,
            gamma: float,
            convention="intrinsic") -> np.ndarray:
    """Rotate to orientation described by zxz euler angles.

    Calculate rotation matrix as determined by zxz Euler angles. Convention can
    be either intrinsic, in which rotations are performed on the moving coordinate
    system XYZ in reference to a fixed coordinate system xyz, or extrinsic, 
    in which the rotations are performed about the fixed coordinate system xyz.
    Intrisic rotations are performed alpha-beta-gamma, while extrinsic rotations
    are perfomed gamma-beta-alpha.

    Args:
        alpha (float): rotation around Z / z in radians
        beta (float): rotation around X'/ x in radians
        gammma (float): rotation around Z''/ z in radians
        convention (float): "intrinsic" or "extrinsic" (default: "intrinsic")

    Returns:
        np.ndarray : 3x3 rotation matrix
    """

    if convention == "intrinsic":
        return scRotation.from_euler("ZXZ", [alpha, beta, gamma],
                                     degrees=False).as_matrix()
    elif convention == "extrinsic":
        return scRotation.from_euler("zxz", [gamma, alpha, beta],
                                     degrees=False).as_matrix()
    else:
        raise (ValueError("Invalid argument for convention."))


def s2s_alignment(M_plane: Plane, T_plane: Plane, M_point: np.ndarray,
                  T_point: np.ndarray) -> MultiTransform:
    """Calculates a transformation aligning two surfaces to eachother.

    Aligns two surfaces, represented by algebraic plane objects, such that 
    their normal vectors are anti-parallel. The translational degree of freedom
    is resolved by minimizing the distance of a point in the target alignment to
    a transformed point from the moving frame. Returned Transformation sequence
    is a rotation followed by translation.

    Specifically, given a point in the moving frame M_p with plane M and a 
    point in the target frame T_p with plane T, calculate rotation R such that 
    (R @ M.normal ) .* T.normal = -1 and translation vector T_t under the 
    constraint T_t = min T_t ||R @ M_p - T_p||

    Args:
        M_plane (Plane): Moving plane to align to another surface
        T_plane (Plane): Target plane to which M_plane is aligned 
        M_point (np.ndarray): Point in moving plane reference frame
        T_point (np.ndarray): Point in target plane reference frame

    Returns:
        MultiTransform: Transformation routine represented by surface alignment.
    """
    R = rot_vtv(M_plane.normal, -1.0 * T_plane.normal)

    R_t = Rotation(matrix=R, origin=M_plane.point)
    M_plane_r = R_t.applyTransformation_alg(M_plane)
    M_point_r = R_t.applyTransformation(M_point)

    proj_point = T_plane.project_point(M_point)
    plane_shift = T_plane.project_point(T_point) - proj_point

    ortho_shift = 1.0 * T_plane.dist_from_plane(M_plane.point) * T_plane.normal
    T_t = Translation(shift=ortho_shift + plane_shift)

    return MultiTransform(transforms=[R_t, T_t])



import numpy as np


class Voxel():
    """Voxel class used to span space for generators and track transformations.

    Voxels provide an alterable view of bases and orientation of crystalline
    generators and are the actual transformed object, not Generators.
    This is in contrst to applying transformations directly to the underlying 
    pymatgen Structure object, for speed and ease of manipulation.
    Voxels also help determine how much of a "block" to a Generator needs to 
    build for the purpose of supplying atoms to a larger volume.

    Attributes:
        scale (float): Scaling factor of basis set.
        bases (np.ndarray): Basis vectors defining crystal unit cell.
        sbases (np.ndarray): Scaled basis set.
        reciprocal_bases (np.ndarray): Basis vectors defining unit cell of reciprocal lattice.
        origin (np.ndarray): Origin of Voxel grid.
    """

    def __init__(self,
                 bases: np.ndarray = np.identity(3),
                 scale: float = np.array([1]),
                 origin: np.ndarray = np.array([0.0, 0.0, 0.0])):
        self._scale = None
        self._bases = None
        self._origin = None
        self.scale = scale
        self.bases = bases
        self.origin = origin

    @property
    def scale(self):
        """Scaling factor of basis set."""
        return self._scale

    @scale.setter
    def scale(self, scale: float):
        try:
            self.bases * np.array(scale)
        except ValueError:
            raise ValueError(
                "Bases and scale are not broadcastable. Scale array must be of shape=(1,) (3,), (1,3), or (3,1)"
            )

        self._scale = np.array(scale)

    @property
    def bases(self):
        """Basis vectors defining crystal unit cell. Vectors are rows of matrix."""
        return self._bases

    @bases.setter
    def bases(self, bases: np.ndarray):
        bases = np.array(bases)

        assert (
            bases.shape == (3, 3)
        ), '''Bases must be 3x3 numpy array that defines vectors that span 3D space
            [0,:] = [x_1, y_1, z_1]
            [1,:] = [x_2, y_2, z_2]
            [2,:] = [x_3, y_3, z_3]'''

        #check for collinearity
        assert (not self._collinear(
            bases[0, :], bases[1, :])), "Bases vectors must linearly indepedent"
        assert (not self._collinear(
            bases[0, :], bases[2, :])), "Bases vectors must linearly indepedent"
        assert (not self._collinear(
            bases[1, :], bases[2, :])), "Bases vectors must linearly indepedent"

        self._bases = bases

    @property
    def origin(self):
        """Origin of Voxel grid in space."""
        return self._origin

    @origin.setter
    def origin(self, origin: np.ndarray):
        assert (origin.shape == (3,)), "Origin must have shape (3,)"
        self._origin = np.array(origin)

    @property
    def sbases(self):
        """Basis vectors defining crystal unit cell, scaled by scaling factor."""
        return self._bases * self._scale

    @property
    def reciprocal_bases(self):
        """Basis vectors defining unit cell of reciprocal lattice."""
        return np.linalg.inv(self.sbases)

    def _collinear(self, vec1: np.ndarray, vec2: np.ndarray):
        """Check if two vectors are collinear. Used for determining validity of basis set.

        Args:
            vec1 (np.ndarray): first vector
            vec2 (np.ndarray): second vector

        Returns:
            bool indicating whether vectors are collinear.
        """
        return np.abs((np.dot(vec1, vec2) /
                       (np.linalg.norm(vec1) * np.linalg.norm(vec2)))) == 1.0

    def get_extents(self, box: np.ndarray):
        """Determine minimum contiguous block of voxels that fully covers a space.

        Args:
            box (np.ndarray): Set of points defining extremities of space.

        Returns:
            Tuple of minimum extents and maximum extents indicating how many
            voxels to tile in space, and where, to span a given region.
        """
        extents = []
        box = box - self.origin
        for point in box:
            extent = np.linalg.solve(self.sbases, np.array(point))
            extents.append(extent)

        extents = np.array(extents)
        min_extent = np.floor(np.min(extents, axis=0))
        max_extent = np.ceil(np.max(extents, axis=0))

        return min_extent.astype(np.int64), max_extent.astype(np.int64)
